@@ -3,8 +3,8 @@
 
 git clone https://boringssl.googlesource.com/boringssl
 cd boringssl
-git checkout 36a41bf0bf2dd3176f8780e09c03585351f29963
+git checkout 31bad2514d21f6207f3925ba56754611c462a873
 mkdir build
 cd build
 cmake -DCMAKE_POSITION_INDEPENDENT_CODE=ON ..
-make -j$(nproc)
+make -j"$(nproc 2> /dev/null || sysctl -n hw.ncpu)"
