@@ -5,5 +5,5 @@ git clone https://github.com/ngtcp2/nghttp3
 cd nghttp3
 autoreconf -i
 ./configure --prefix=$PWD/build --enable-lib-only
-make -j$(nproc) check
+make -j"$(nproc 2> /dev/null || sysctl -n hw.ncpu)" check
 make install
