@@ -55,6 +55,10 @@ namespace {
 auto randgen = util::make_mt19937();
 } // namespace
 
+namespace {
+constexpr size_t max_preferred_versionslen = 4;
+} // namespace
+
 Config config{};
 
 Stream::Stream(const Request &req, int64_t stream_id)
@@ -73,10 +77,10 @@ int Stream::open_file(const std::string_view &path) {
 
   auto it = std::find(std::rbegin(path), std::rend(path), '/').base();
   if (it == std::end(path)) {
-    filename = "index.html";
+    filename = "index.html"sv;
   } else {
     filename = std::string_view{it, static_cast<size_t>(std::end(path) - it)};
-    if (filename == ".." || filename == ".") {
+    if (filename == ".."sv || filename == "."sv) {
       std::cerr << "Invalid file name: " << filename << std::endl;
       return -1;
     }
@@ -120,23 +124,6 @@ void readcb(struct ev_loop *loop, ev_io *w, int revents) {
 
 namespace {
 void timeoutcb(struct ev_loop *loop, ev_timer *w, int revents) {
-  auto c = static_cast<Client *>(w->data);
-
-  if (!config.quiet) {
-    std::cerr << "Timeout" << std::endl;
-  }
-
-  c->idle_timeout();
-}
-} // namespace
-
-void Client::idle_timeout() {
-  last_error_ = quic_err_idle_timeout();
-  disconnect();
-}
-
-namespace {
-void retransmitcb(struct ev_loop *loop, ev_timer *w, int revents) {
   int rv;
   auto c = static_cast<Client *>(w->data);
 
@@ -183,7 +170,8 @@ void siginthandler(struct ev_loop *loop, ev_signal *w, int revents) {
 }
 } // namespace
 
-Client::Client(struct ev_loop *loop)
+Client::Client(struct ev_loop *loop, uint32_t client_chosen_version,
+               uint32_t original_version)
     : remote_addr_{},
       loop_(loop),
       httpconn_(nullptr),
@@ -192,18 +180,17 @@ Client::Client(struct ev_loop *loop)
       nstreams_done_(0),
       nstreams_closed_(0),
       nkey_update_(0),
-      version_(0),
+      client_chosen_version_(client_chosen_version),
+      original_version_(original_version),
       early_data_(false),
       should_exit_(false),
+      should_exit_on_handshake_confirmed_(false),
       handshake_confirmed_(false),
       tx_{} {
   ev_io_init(&wev_, writecb, 0, EV_WRITE);
   wev_.data = this;
-  ev_timer_init(&timer_, timeoutcb, 0.,
-                static_cast<double>(config.timeout) / NGTCP2_SECONDS);
+  ev_timer_init(&timer_, timeoutcb, 0., 0.);
   timer_.data = this;
-  ev_timer_init(&rttimer_, retransmitcb, 0., 0.);
-  rttimer_.data = this;
   ev_timer_init(&change_local_addr_timer_, change_local_addrcb,
                 static_cast<double>(config.change_local_addr) / NGTCP2_SECONDS,
                 0.);
@@ -215,8 +202,6 @@ Client::Client(struct ev_loop *loop)
                 static_cast<double>(config.delay_stream) / NGTCP2_SECONDS, 0.);
   delay_stream_timer_.data = this;
   ev_signal_init(&sigintev_, siginthandler, SIGINT);
-
-  application_rx_key_cb_ = [this]() { return setup_httpconn(); };
 }
 
 Client::~Client() {
@@ -238,7 +223,6 @@ void Client::disconnect() {
   ev_timer_stop(loop_, &delay_stream_timer_);
   ev_timer_stop(loop_, &key_update_timer_);
   ev_timer_stop(loop_, &change_local_addr_timer_);
-  ev_timer_stop(loop_, &rttimer_);
   ev_timer_stop(loop_, &timer_);
 
   ev_io_stop(loop_, &wev_);
@@ -318,18 +302,16 @@ int Client::handshake_completed() {
       std::cerr << "Early data was rejected by server" << std::endl;
     }
 
+    // Some TLS backends only report early data rejection after
+    // handshake completion (e.g., OpenSSL).  For TLS backends which
+    // report it early (e.g., BoringSSL and PicoTLS), the following
+    // functions are noop.
     if (auto rv = ngtcp2_conn_early_data_rejected(conn_); rv != 0) {
       std::cerr << "ngtcp2_conn_early_data_rejected: " << ngtcp2_strerror(rv)
                 << std::endl;
       return -1;
     }
 
-    nghttp3_conn_del(httpconn_);
-    httpconn_ = nullptr;
-
-    nstreams_done_ = 0;
-    streams_.clear();
-
     if (setup_httpconn() != 0) {
       return -1;
     }
@@ -342,6 +324,15 @@ int Client::handshake_completed() {
               << std::endl;
   }
 
+  if (config.tp_file) {
+    auto params = ngtcp2_conn_get_remote_transport_params(conn_);
+
+    if (write_transport_params(config.tp_file, params) != 0) {
+      std::cerr << "Could not write transport parameters in " << config.tp_file
+                << std::endl;
+    }
+  }
+
   return 0;
 }
 
@@ -374,9 +365,29 @@ int Client::handshake_confirmed() {
     start_delay_stream_timer();
   }
 
+  if (should_exit_on_handshake_confirmed_) {
+    should_exit_ = true;
+  }
+
   return 0;
 }
 
+namespace {
+int recv_version_negotiation(ngtcp2_conn *conn, const ngtcp2_pkt_hd *hd,
+                             const uint32_t *sv, size_t nsv, void *user_data) {
+  auto c = static_cast<Client *>(user_data);
+
+  c->recv_version_negotiation(sv, nsv);
+
+  return 0;
+}
+} // namespace
+
+void Client::recv_version_negotiation(const uint32_t *sv, size_t nsv) {
+  offered_versions_.resize(nsv);
+  std::copy_n(sv, nsv, std::begin(offered_versions_));
+}
+
 namespace {
 int stream_close(ngtcp2_conn *conn, uint32_t flags, int64_t stream_id,
                  uint64_t app_error_code, void *user_data,
@@ -461,13 +472,6 @@ int get_new_connection_id(ngtcp2_conn *conn, ngtcp2_cid *cid, uint8_t *token,
 }
 } // namespace
 
-namespace {
-int remove_connection_id(ngtcp2_conn *conn, const ngtcp2_cid *cid,
-                         void *user_data) {
-  return 0;
-}
-} // namespace
-
 namespace {
 int do_hp_mask(uint8_t *dest, const ngtcp2_crypto_cipher *hp,
                const ngtcp2_crypto_cipher_ctx *hp_ctx, const uint8_t *sample) {
@@ -586,8 +590,42 @@ int recv_new_token(ngtcp2_conn *conn, const ngtcp2_vec *token,
 }
 } // namespace
 
+namespace {
+int recv_rx_key(ngtcp2_conn *conn, ngtcp2_crypto_level level, void *user_data) {
+  if (level != NGTCP2_CRYPTO_LEVEL_APPLICATION) {
+    return 0;
+  }
+
+  auto c = static_cast<Client *>(user_data);
+  if ((!c->get_early_data() || ngtcp2_conn_get_early_data_rejected(conn)) &&
+      c->setup_httpconn() != 0) {
+    return NGTCP2_ERR_CALLBACK_FAILURE;
+  }
+
+  return 0;
+}
+} // namespace
+
+namespace {
+int early_data_rejected(ngtcp2_conn *conn, void *user_data) {
+  auto c = static_cast<Client *>(user_data);
+
+  c->early_data_rejected();
+
+  return 0;
+}
+} // namespace
+
+void Client::early_data_rejected() {
+  nghttp3_conn_del(httpconn_);
+  httpconn_ = nullptr;
+
+  nstreams_done_ = 0;
+  streams_.clear();
+}
+
 int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
-                 const char *addr, const char *port, uint32_t version,
+                 const char *addr, const char *port,
                  TLSClientContext &tls_ctx) {
   endpoints_.reserve(4);
 
@@ -602,14 +640,13 @@ int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
   remote_addr_ = remote_addr;
   addr_ = addr;
   port_ = port;
-  version_ = version;
 
   auto callbacks = ngtcp2_callbacks{
       ngtcp2_crypto_client_initial_cb,
       nullptr, // recv_client_initial
       ::recv_crypto_data,
       ::handshake_completed,
-      nullptr, // recv_version_negotiation
+      ::recv_version_negotiation,
       ngtcp2_crypto_encrypt_cb,
       ngtcp2_crypto_decrypt_cb,
       do_hp_mask,
@@ -623,7 +660,7 @@ int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
       nullptr, // extend_max_streams_uni
       rand,
       get_new_connection_id,
-      remove_connection_id,
+      nullptr, // remove_connection_id
       ::update_key,
       path_validation,
       ::select_preferred_address,
@@ -641,6 +678,10 @@ int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
       nullptr, // lost_datagram
       ngtcp2_crypto_get_path_challenge_data_cb,
       stream_stop_sending,
+      ngtcp2_crypto_version_negotiation_cb,
+      ::recv_rx_key,
+      nullptr, // recv_tx_key
+      ::early_data_rejected,
   };
 
   ngtcp2_cid scid, dcid;
@@ -690,9 +731,13 @@ int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
   settings.initial_rtt = config.initial_rtt;
   settings.max_window = config.max_window;
   settings.max_stream_window = config.max_stream_window;
-  settings.max_udp_payload_size = config.max_udp_payload_size;
-  settings.no_udp_payload_size_shaping = 1;
+  if (config.max_udp_payload_size) {
+    settings.max_tx_udp_payload_size = config.max_udp_payload_size;
+    settings.no_tx_udp_payload_size_shaping = 1;
+  }
   settings.handshake_timeout = config.handshake_timeout;
+  settings.no_pmtud = config.no_pmtud;
+  settings.ack_thresh = config.ack_thresh;
 
   std::string token;
 
@@ -707,6 +752,18 @@ int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
     }
   }
 
+  if (!config.other_versions.empty()) {
+    settings.other_versions = config.other_versions.data();
+    settings.other_versionslen = config.other_versions.size();
+  }
+
+  if (!config.preferred_versions.empty()) {
+    settings.preferred_versions = config.preferred_versions.data();
+    settings.preferred_versionslen = config.preferred_versions.size();
+  }
+
+  settings.original_version = original_version_;
+
   ngtcp2_transport_params params;
   ngtcp2_transport_params_default(&params);
   params.initial_max_stream_data_bidi_local = config.max_stream_data_bidi_local;
@@ -730,17 +787,17 @@ int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
       },
       &ep,
   };
-  auto rv =
-      ngtcp2_conn_client_new(&conn_, &dcid, &scid, &path, version, &callbacks,
-                             &settings, &params, nullptr, this);
+  auto rv = ngtcp2_conn_client_new(&conn_, &dcid, &scid, &path,
+                                   client_chosen_version_, &callbacks,
+                                   &settings, &params, nullptr, this);
 
   if (rv != 0) {
     std::cerr << "ngtcp2_conn_client_new: " << ngtcp2_strerror(rv) << std::endl;
     return -1;
   }
 
-  if (tls_session_.init(early_data_, tls_ctx, addr_, this, version,
-                        AppProtocol::H3) != 0) {
+  if (tls_session_.init(early_data_, tls_ctx, addr_, this,
+                        client_chosen_version_, AppProtocol::H3) != 0) {
     return -1;
   }
 
@@ -761,7 +818,6 @@ int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
   }
 
   ev_io_start(loop_, &ep.rev);
-  ev_timer_again(loop_, &timer_);
 
   ev_signal_start(loop_, &sigintev_);
 
@@ -786,24 +842,13 @@ int Client::feed_data(const Endpoint &ep, const sockaddr *sa, socklen_t salen,
                                      util::timestamp(loop_));
       rv != 0) {
     std::cerr << "ngtcp2_conn_read_pkt: " << ngtcp2_strerror(rv) << std::endl;
-    switch (rv) {
-    case NGTCP2_ERR_REQUIRED_TRANSPORT_PARAM:
-    case NGTCP2_ERR_MALFORMED_TRANSPORT_PARAM:
-    case NGTCP2_ERR_TRANSPORT_PARAM:
-    case NGTCP2_ERR_PROTO: // with failed TP validation, we get this.
-      // If rv indicates transport_parameters related error, we should
-      // send TRANSPORT_PARAMETER_ERROR even if last_error_.code is
-      // already set.  This is because OpenSSL might set Alert.
-      last_error_ = quic_err_transport(rv);
-      break;
-    case NGTCP2_ERR_CRYPTO:
-      if (!last_error_.code) {
-        process_unhandled_tls_alert();
-      }
-      // fall through
-    default:
-      if (!last_error_.code) {
-        last_error_ = quic_err_transport(rv);
+    if (!last_error_.error_code) {
+      if (rv == NGTCP2_ERR_CRYPTO) {
+        ngtcp2_connection_close_error_set_transport_error_tls_alert(
+            &last_error_, ngtcp2_conn_get_tls_alert(conn_), nullptr, 0);
+      } else {
+        ngtcp2_connection_close_error_set_transport_error_liberr(
+            &last_error_, rv, nullptr, 0);
       }
     }
     disconnect();
@@ -813,7 +858,7 @@ int Client::feed_data(const Endpoint &ep, const sockaddr *sa, socklen_t salen,
 }
 
 int Client::on_read(const Endpoint &ep) {
-  std::array<uint8_t, 65536> buf;
+  std::array<uint8_t, 64_k> buf;
   sockaddr_union su;
   size_t pktcnt = 0;
   ngtcp2_pkt_info pi;
@@ -870,38 +915,24 @@ int Client::on_read(const Endpoint &ep) {
   }
 
   if (should_exit_) {
-    last_error_ = quic_err_app(0);
+    ngtcp2_connection_close_error_set_application_error(
+        &last_error_, nghttp3_err_infer_quic_app_error_code(0), nullptr, 0);
     disconnect();
     return -1;
   }
 
-  reset_idle_timer();
+  update_timer();
 
   return 0;
 }
 
-void Client::reset_idle_timer() {
-  auto now = util::timestamp(loop_);
-  auto idle_expiry = ngtcp2_conn_get_idle_expiry(conn_);
-  timer_.repeat =
-      idle_expiry > now
-          ? static_cast<ev_tstamp>(idle_expiry - now) / NGTCP2_SECONDS
-          : 1e-9;
-
-  if (!config.quiet) {
-    std::cerr << "Set idle timer=" << std::fixed << timer_.repeat << "s"
-              << std::defaultfloat << std::endl;
-  }
-
-  ev_timer_again(loop_, &timer_);
-}
-
 int Client::handle_expiry() {
   auto now = util::timestamp(loop_);
   if (auto rv = ngtcp2_conn_handle_expiry(conn_, now); rv != 0) {
     std::cerr << "ngtcp2_conn_handle_expiry: " << ngtcp2_strerror(rv)
               << std::endl;
-    last_error_ = quic_err_transport(NGTCP2_ERR_INTERNAL);
+    ngtcp2_connection_close_error_set_transport_error_liberr(&last_error_, rv,
+                                                             nullptr, 0);
     disconnect();
     return -1;
   }
@@ -925,12 +956,13 @@ int Client::on_write() {
   }
 
   if (should_exit_) {
-    last_error_ = quic_err_app(0);
+    ngtcp2_connection_close_error_set_application_error(
+        &last_error_, nghttp3_err_infer_quic_app_error_code(0), nullptr, 0);
     disconnect();
     return -1;
   }
 
-  schedule_retransmit();
+  update_timer();
   return 0;
 }
 
@@ -938,12 +970,8 @@ int Client::write_streams() {
   std::array<nghttp3_vec, 16> vec;
   ngtcp2_path_storage ps;
   size_t pktcnt = 0;
-  auto max_udp_payload_size = ngtcp2_conn_get_path_max_udp_payload_size(conn_);
-  size_t max_pktcnt =
-      (config.cc_algo == NGTCP2_CC_ALGO_BBR ||
-       config.cc_algo == NGTCP2_CC_ALGO_BBR2)
-          ? ngtcp2_conn_get_send_quantum(conn_) / max_udp_payload_size
-          : 10;
+  auto max_udp_payload_size = ngtcp2_conn_get_max_tx_udp_payload_size(conn_);
+  auto max_pktcnt = ngtcp2_conn_get_send_quantum(conn_) / max_udp_payload_size;
   auto ts = util::timestamp(loop_);
 
   ngtcp2_path_storage_zero(&ps);
@@ -959,7 +987,9 @@ int Client::write_streams() {
       if (sveccnt < 0) {
         std::cerr << "nghttp3_conn_writev_stream: " << nghttp3_strerror(sveccnt)
                   << std::endl;
-        last_error_ = quic_err_app(sveccnt);
+        ngtcp2_connection_close_error_set_application_error(
+            &last_error_, nghttp3_err_infer_quic_app_error_code(sveccnt),
+            nullptr, 0);
         disconnect();
         return -1;
       }
@@ -983,25 +1013,11 @@ int Client::write_streams() {
       switch (nwrite) {
       case NGTCP2_ERR_STREAM_DATA_BLOCKED:
         assert(ndatalen == -1);
-        if (auto rv = nghttp3_conn_block_stream(httpconn_, stream_id);
-            rv != 0) {
-          std::cerr << "nghttp3_conn_block_stream: " << nghttp3_strerror(rv)
-                    << std::endl;
-          last_error_ = quic_err_app(rv);
-          disconnect();
-          return -1;
-        }
+        nghttp3_conn_block_stream(httpconn_, stream_id);
         continue;
       case NGTCP2_ERR_STREAM_SHUT_WR:
         assert(ndatalen == -1);
-        if (auto rv = nghttp3_conn_shutdown_stream_write(httpconn_, stream_id);
-            rv != 0) {
-          std::cerr << "nghttp3_conn_shutdown_stream_write: "
-                    << nghttp3_strerror(rv) << std::endl;
-          last_error_ = quic_err_app(rv);
-          disconnect();
-          return -1;
-        }
+        nghttp3_conn_shutdown_stream_write(httpconn_, stream_id);
         continue;
       case NGTCP2_ERR_WRITE_MORE:
         assert(ndatalen >= 0);
@@ -1010,7 +1026,9 @@ int Client::write_streams() {
             rv != 0) {
           std::cerr << "nghttp3_conn_add_write_offset: " << nghttp3_strerror(rv)
                     << std::endl;
-          last_error_ = quic_err_app(rv);
+          ngtcp2_connection_close_error_set_application_error(
+              &last_error_, nghttp3_err_infer_quic_app_error_code(rv), nullptr,
+              0);
           disconnect();
           return -1;
         }
@@ -1021,7 +1039,8 @@ int Client::write_streams() {
 
       std::cerr << "ngtcp2_conn_write_stream: " << ngtcp2_strerror(nwrite)
                 << std::endl;
-      last_error_ = quic_err_transport(nwrite);
+      ngtcp2_connection_close_error_set_transport_error_liberr(
+          &last_error_, nwrite, nullptr, 0);
       disconnect();
       return -1;
     } else if (ndatalen >= 0) {
@@ -1030,7 +1049,9 @@ int Client::write_streams() {
           rv != 0) {
         std::cerr << "nghttp3_conn_add_write_offset: " << nghttp3_strerror(rv)
                   << std::endl;
-        last_error_ = quic_err_app(rv);
+        ngtcp2_connection_close_error_set_application_error(
+            &last_error_, nghttp3_err_infer_quic_app_error_code(rv), nullptr,
+            0);
         disconnect();
         return -1;
       }
@@ -1043,15 +1064,14 @@ int Client::write_streams() {
       return 0;
     }
 
-    reset_idle_timer();
-
     auto &ep = *static_cast<Endpoint *>(ps.path.user_data);
 
     if (auto rv =
             send_packet(ep, ps.path.remote, pi.ecn, tx_.data.data(), nwrite);
         rv != NETWORK_ERR_OK) {
       if (rv != NETWORK_ERR_SEND_BLOCKED) {
-        last_error_ = quic_err_transport(NGTCP2_ERR_INTERNAL);
+        ngtcp2_connection_close_error_set_transport_error_liberr(
+            &last_error_, NGTCP2_ERR_INTERNAL, nullptr, 0);
         disconnect();
 
         return rv;
@@ -1071,17 +1091,28 @@ int Client::write_streams() {
   }
 }
 
-void Client::schedule_retransmit() {
+void Client::update_timer() {
   auto expiry = ngtcp2_conn_get_expiry(conn_);
   auto now = util::timestamp(loop_);
-  auto t = expiry < now ? 1e-9
-                        : static_cast<ev_tstamp>(expiry - now) / NGTCP2_SECONDS;
+
+  if (expiry <= now) {
+    if (!config.quiet) {
+      auto t = static_cast<ev_tstamp>(now - expiry) / NGTCP2_SECONDS;
+      std::cerr << "Timer has already expired: " << t << "s" << std::endl;
+    }
+
+    ev_feed_event(loop_, &timer_, EV_TIMER);
+
+    return;
+  }
+
+  auto t = static_cast<ev_tstamp>(expiry - now) / NGTCP2_SECONDS;
   if (!config.quiet) {
     std::cerr << "Set timer=" << std::fixed << t << "s" << std::defaultfloat
               << std::endl;
   }
-  rttimer_.repeat = t;
-  ev_timer_again(loop_, &rttimer_);
+  timer_.repeat = t;
+  ev_timer_again(loop_, &timer_);
 }
 
 #ifdef HAVE_LINUX_RTNETLINK_H
@@ -1168,14 +1199,9 @@ int udp_sock(int family) {
     return -1;
   }
 
-  auto val = 1;
-  if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &val,
-                 static_cast<socklen_t>(sizeof(val))) == -1) {
-    close(fd);
-    return -1;
-  }
-
   fd_set_recv_ecn(fd, family);
+  fd_set_ip_mtu_discover(fd, family);
+  fd_set_ip_dontfrag(fd, family);
 
   return fd;
 }
@@ -1439,6 +1465,9 @@ int Client::send_packet(const Endpoint &ep, const ngtcp2_addr &remote_addr,
       return NETWORK_ERR_SEND_BLOCKED;
     }
     std::cerr << "sendmsg: " << strerror(errno) << std::endl;
+    if (errno == EMSGSIZE) {
+      return 0;
+    }
     return NETWORK_ERR_FATAL;
   }
 
@@ -1503,7 +1532,8 @@ int Client::send_blocked_packet() {
       return 0;
     }
 
-    last_error_ = quic_err_transport(NGTCP2_ERR_INTERNAL);
+    ngtcp2_connection_close_error_set_transport_error_liberr(
+        &last_error_, NGTCP2_ERR_INTERNAL, nullptr, 0);
     disconnect();
 
     return rv;
@@ -1515,12 +1545,8 @@ int Client::send_blocked_packet() {
 }
 
 int Client::handle_error() {
-  if (!conn_ || ngtcp2_conn_is_in_closing_period(conn_)) {
-    return 0;
-  }
-
-  if (last_error_.type == QUICErrorType::TransportVersionNegotiation ||
-      last_error_.type == QUICErrorType::TransportIdleTimeout) {
+  if (!conn_ || ngtcp2_conn_is_in_closing_period(conn_) ||
+      ngtcp2_conn_is_in_draining_period(conn_)) {
     return 0;
   }
 
@@ -1531,25 +1557,18 @@ int Client::handle_error() {
   ngtcp2_path_storage_zero(&ps);
 
   ngtcp2_pkt_info pi;
-  ngtcp2_ssize nwrite;
-  if (last_error_.type == QUICErrorType::Transport) {
-    nwrite = ngtcp2_conn_write_connection_close(
-        conn_, &ps.path, &pi, buf.data(), buf.size(), last_error_.code, nullptr,
-        0, util::timestamp(loop_));
-    if (nwrite < 0) {
-      std::cerr << "ngtcp2_conn_write_connection_close: "
-                << ngtcp2_strerror(nwrite) << std::endl;
-      return -1;
-    }
-  } else {
-    nwrite = ngtcp2_conn_write_application_close(
-        conn_, &ps.path, &pi, buf.data(), buf.size(), last_error_.code, nullptr,
-        0, util::timestamp(loop_));
-    if (nwrite < 0) {
-      std::cerr << "ngtcp2_conn_write_application_close: "
-                << ngtcp2_strerror(nwrite) << std::endl;
-      return -1;
-    }
+
+  auto nwrite = ngtcp2_conn_write_connection_close(
+      conn_, &ps.path, &pi, buf.data(), buf.size(), &last_error_,
+      util::timestamp(loop_));
+  if (nwrite < 0) {
+    std::cerr << "ngtcp2_conn_write_connection_close: "
+              << ngtcp2_strerror(nwrite) << std::endl;
+    return -1;
+  }
+
+  if (nwrite == 0) {
+    return 0;
   }
 
   return send_packet(*static_cast<Endpoint *>(ps.path.user_data),
@@ -1577,7 +1596,8 @@ int Client::on_stream_close(int64_t stream_id, uint64_t app_error_code) {
     default:
       std::cerr << "nghttp3_conn_close_stream: " << nghttp3_strerror(rv)
                 << std::endl;
-      last_error_ = quic_err_app(rv);
+      ngtcp2_connection_close_error_set_application_error(
+          &last_error_, nghttp3_err_infer_quic_app_error_code(rv), nullptr, 0);
       return -1;
     }
   }
@@ -1668,16 +1688,16 @@ int Client::submit_http_request(const Stream *stream) {
   const auto &req = stream->req;
 
   std::array<nghttp3_nv, 6> nva{
-      util::make_nv(":method", config.http_method),
-      util::make_nv(":scheme", req.scheme),
-      util::make_nv(":authority", req.authority),
-      util::make_nv(":path", req.path),
-      util::make_nv("user-agent", "nghttp3/ngtcp2 client"),
+      util::make_nv_nn(":method", config.http_method),
+      util::make_nv_nn(":scheme", req.scheme),
+      util::make_nv_nn(":authority", req.authority),
+      util::make_nv_nn(":path", req.path),
+      util::make_nv_nn("user-agent", "nghttp3/ngtcp2 client"),
   };
   size_t nvlen = 5;
   if (config.fd != -1) {
-    content_length_str = std::to_string(config.datalen);
-    nva[nvlen++] = util::make_nv("content-length", content_length_str);
+    content_length_str = util::format_uint(config.datalen);
+    nva[nvlen++] = util::make_nv_nc("content-length", content_length_str);
   }
 
   if (!config.quiet) {
@@ -1706,7 +1726,9 @@ int Client::recv_stream_data(uint32_t flags, int64_t stream_id,
   if (nconsumed < 0) {
     std::cerr << "nghttp3_conn_read_stream: " << nghttp3_strerror(nconsumed)
               << std::endl;
-    last_error_ = quic_err_app(nconsumed);
+    ngtcp2_connection_close_error_set_application_error(
+        &last_error_, nghttp3_err_infer_quic_app_error_code(nconsumed), nullptr,
+        0);
     return -1;
   }
 
@@ -1729,10 +1751,6 @@ int Client::acked_stream_data_offset(int64_t stream_id, uint64_t datalen) {
 
 int Client::select_preferred_address(Address &selected_addr,
                                      const ngtcp2_preferred_addr *paddr) {
-  int af;
-  const uint8_t *binaddr;
-  uint16_t port;
-
   auto path = ngtcp2_conn_get_path(conn_);
 
   switch (path->local.addr->sa_family) {
@@ -1740,53 +1758,34 @@ int Client::select_preferred_address(Address &selected_addr,
     if (!paddr->ipv4_present) {
       return -1;
     }
-    af = AF_INET;
-    binaddr = paddr->ipv4_addr;
-    port = paddr->ipv4_port;
+    selected_addr.su.in = paddr->ipv4;
+    selected_addr.len = sizeof(paddr->ipv4);
     break;
   case AF_INET6:
     if (!paddr->ipv6_present) {
       return -1;
     }
-    af = AF_INET6;
-    binaddr = paddr->ipv6_addr;
-    port = paddr->ipv6_port;
+    selected_addr.su.in6 = paddr->ipv6;
+    selected_addr.len = sizeof(paddr->ipv6);
     break;
   default:
     return -1;
   }
 
-  char host[NI_MAXHOST];
-  if (inet_ntop(af, binaddr, host, sizeof(host)) == nullptr) {
-    std::cerr << "inet_ntop: " << strerror(errno) << std::endl;
+  char host[NI_MAXHOST], service[NI_MAXSERV];
+  if (auto rv = getnameinfo(&selected_addr.su.sa, selected_addr.len, host,
+                            sizeof(host), service, sizeof(service),
+                            NI_NUMERICHOST | NI_NUMERICSERV);
+      rv != 0) {
+    std::cerr << "getnameinfo: " << gai_strerror(rv) << std::endl;
     return -1;
   }
 
   if (!config.quiet) {
     std::cerr << "selected server preferred_address is [" << host
-              << "]:" << port << std::endl;
+              << "]:" << service << std::endl;
   }
 
-  addrinfo hints{};
-  addrinfo *res;
-
-  hints.ai_flags = AI_NUMERICHOST | AI_NUMERICSERV;
-  hints.ai_family = af;
-  hints.ai_socktype = SOCK_DGRAM;
-
-  if (auto rv = getaddrinfo(host, std::to_string(port).c_str(), &hints, &res);
-      rv != 0) {
-    std::cerr << "getaddrinfo: " << gai_strerror(rv) << std::endl;
-    return -1;
-  }
-
-  assert(res);
-
-  selected_addr.len = res->ai_addrlen;
-  memcpy(&selected_addr.su, res->ai_addr, res->ai_addrlen);
-
-  freeaddrinfo(res);
-
   return 0;
 }
 
@@ -1917,9 +1916,29 @@ int Client::stop_sending(int64_t stream_id, uint64_t app_error_code) {
       rv != 0) {
     std::cerr << "ngtcp2_conn_shutdown_stream_read: " << ngtcp2_strerror(rv)
               << std::endl;
-    if (rv == NGTCP2_ERR_STREAM_NOT_FOUND) {
-      return 0;
-    }
+    return -1;
+  }
+  return 0;
+}
+
+namespace {
+int http_reset_stream(nghttp3_conn *conn, int64_t stream_id,
+                      uint64_t app_error_code, void *user_data,
+                      void *stream_user_data) {
+  auto c = static_cast<Client *>(user_data);
+  if (c->reset_stream(stream_id, app_error_code) != 0) {
+    return NGHTTP3_ERR_CALLBACK_FAILURE;
+  }
+  return 0;
+}
+} // namespace
+
+int Client::reset_stream(int64_t stream_id, uint64_t app_error_code) {
+  if (auto rv =
+          ngtcp2_conn_shutdown_stream_write(conn_, stream_id, app_error_code);
+      rv != 0) {
+    std::cerr << "ngtcp2_conn_shutdown_stream_write: " << ngtcp2_strerror(rv)
+              << std::endl;
     return -1;
   }
   return 0;
@@ -1947,7 +1966,11 @@ int Client::http_stream_close(int64_t stream_id, uint64_t app_error_code) {
         (config.exit_on_all_streams_close &&
          config.nstreams == nstreams_done_ &&
          nstreams_closed_ == nstreams_done_)) {
-      should_exit_ = true;
+      if (handshake_confirmed_) {
+        should_exit_ = true;
+      } else {
+        should_exit_on_handshake_confirmed_ = true;
+      }
     }
   } else {
     assert(!ngtcp2_conn_is_local_stream(conn_, stream_id));
@@ -1977,7 +2000,7 @@ int Client::setup_httpconn() {
   }
 
   nghttp3_callbacks callbacks{
-      nullptr, /* acked_stream_data*/
+      nullptr, // acked_stream_data
       ::http_stream_close,
       ::http_recv_data,
       ::http_deferred_consume,
@@ -1988,10 +2011,13 @@ int Client::setup_httpconn() {
       ::http_recv_trailer,
       ::http_end_trailers,
       ::http_stop_sending,
+      nullptr, // end_stream
+      ::http_reset_stream,
+      nullptr, // shutdown
   };
   nghttp3_settings settings;
   nghttp3_settings_default(&settings);
-  settings.qpack_max_dtable_capacity = 4096;
+  settings.qpack_max_dtable_capacity = 4_k;
   settings.qpack_blocked_streams = 100;
 
   auto mem = nghttp3_mem_default();
@@ -2059,6 +2085,12 @@ int Client::setup_httpconn() {
   return 0;
 }
 
+const std::vector<uint32_t> &Client::get_offered_versions() const {
+  return offered_versions_;
+}
+
+bool Client::get_early_data() const { return early_data_; };
+
 namespace {
 int run(Client &c, const char *addr, const char *port,
         TLSClientContext &tls_ctx) {
@@ -2089,8 +2121,7 @@ int run(Client &c, const char *addr, const char *port,
   }
 #endif // !HAVE_LINUX_RTNETLINK_H
 
-  if (c.init(fd, local_addr, remote_addr, addr, port, config.version,
-             tls_ctx) != 0) {
+  if (c.init(fd, local_addr, remote_addr, addr, port, tls_ctx) != 0) {
     return -1;
   }
 
@@ -2129,7 +2160,7 @@ int parse_uri(Request &req, const char *uri) {
   req.scheme = get_string(uri, u, UF_SCHEMA);
 
   req.authority = get_string(uri, u, UF_HOST);
-  if (util::numeric_host(req.authority.c_str())) {
+  if (util::numeric_host(req.authority.c_str(), AF_INET6)) {
     req.authority = '[' + req.authority + ']';
   }
   if (u.field_set & (1 << UF_PORT)) {
@@ -2189,17 +2220,17 @@ void config_set_default(Config &config) {
   config.version = NGTCP2_PROTO_VER_V1;
   config.timeout = 30 * NGTCP2_SECONDS;
   config.http_method = "GET"sv;
-  config.max_data = 1_m;
-  config.max_stream_data_bidi_local = 256_k;
-  config.max_stream_data_bidi_remote = 256_k;
-  config.max_stream_data_uni = 256_k;
-  config.max_window = 6_m;
-  config.max_stream_window = 6_m;
+  config.max_data = 15_m;
+  config.max_stream_data_bidi_local = 6_m;
+  config.max_stream_data_bidi_remote = 6_m;
+  config.max_stream_data_uni = 6_m;
+  config.max_window = 24_m;
+  config.max_stream_window = 16_m;
   config.max_streams_uni = 100;
   config.cc_algo = NGTCP2_CC_ALGO_CUBIC;
   config.initial_rtt = NGTCP2_DEFAULT_INITIAL_RTT;
-  config.max_udp_payload_size = client_max_udp_payload_size;
   config.handshake_timeout = NGTCP2_DEFAULT_HANDSHAKE_TIMEOUT;
+  config.ack_thresh = 2;
 }
 } // namespace
 
@@ -2231,9 +2262,30 @@ Options:
               list  is  less than  <N>,  <URI>  list is  wrapped.   It
               defaults to 0 which means the number of <URI> specified.
   -v, --version=<HEX>
-              Specify QUIC version to use in hex string.
+              Specify QUIC version to use in hex string.  If the given
+              version is  not supported by libngtcp2,  client will use
+              QUIC v1  long packet  types.  Instead of  specifying hex
+              string,  there  are   special  aliases  available:  "v1"
+              indicates  QUIC  v1,  and "v2draft"  indicates  QUIC  v2
+              draft.
               Default: )"
             << std::hex << "0x" << config.version << std::dec << R"(
+  --preferred-versions=<HEX>[[,<HEX>]...]
+              Specify  QUIC versions  in hex  string in  the order  of
+              preference.   Client chooses  one of  those versions  if
+              client received Version  Negotiation packet from server.
+              These versions must be  supported by libngtcp2.  Instead
+              of  specifying hex  string,  there  are special  aliases
+              available:  "v1"   indicates  QUIC  v1,   and  "v2draft"
+              indicates QUIC v2 draft.
+  --other-versions=<HEX>[[,<HEX>]...]
+              Specify QUIC  versions in  hex string  that are  sent in
+              other_versions  field  of version_information  transport
+              parameter.  This list can include a version which is not
+              supported  by  libngtcp2.   Instead  of  specifying  hex
+              string,  there  are   special  aliases  available:  "v1"
+              indicates  QUIC  v1,  and "v2draft"  indicates  QUIC  v2
+              draft.
   -q, --quiet Suppress debug output.
   -s, --show-secret
               Print out secrets unless --quiet is used.
@@ -2367,12 +2419,16 @@ Options:
             << util::format_uint_iec(config.max_stream_window) << R"(
   --max-udp-payload-size=<SIZE>
               Override maximum UDP payload size that client transmits.
-              Default: )"
-            << config.max_udp_payload_size << R"(
   --handshake-timeout=<DURATION>
               Set the QUIC handshake timeout.
               Default: )"
             << util::format_duration(config.handshake_timeout) << R"(
+  --no-pmtud  Disables Path MTU Discovery.
+  --ack-thresh=<N>
+              The minimum number of the received ACK eliciting packets
+              that triggers immediate acknowledgement.
+              Default: )"
+            << config.ack_thresh << R"(
   -h, --help  Display this help and exit.
 
 ---
@@ -2384,7 +2440,11 @@ Options:
   is 1 second and 500ms is 500  milliseconds).  Units are h, m, s, ms,
   us, or ns (hours,  minutes, seconds, milliseconds, microseconds, and
   nanoseconds respectively).  If  a unit is omitted, a  second is used
-  as unit.)" << std::endl;
+  as unit.
+
+  The  <HEX> argument  is an  hex string  which must  start with  "0x"
+  (e.g., 0x00000001).)"
+            << std::endl;
 }
 } // namespace
 
@@ -2442,6 +2502,10 @@ int main(int argc, char **argv) {
         {"scid", required_argument, &flag, 34},
         {"max-udp-payload-size", required_argument, &flag, 35},
         {"handshake-timeout", required_argument, &flag, 36},
+        {"other-versions", required_argument, &flag, 37},
+        {"no-pmtud", no_argument, &flag, 38},
+        {"preferred-versions", required_argument, &flag, 39},
+        {"ack-thresh", required_argument, &flag, 40},
         {nullptr, 0, nullptr, 0},
     };
 
@@ -2492,10 +2556,25 @@ int main(int argc, char **argv) {
       // --tx-loss
       config.tx_loss_prob = strtod(optarg, nullptr);
       break;
-    case 'v':
+    case 'v': {
       // --version
-      config.version = strtol(optarg, nullptr, 16);
+      if (optarg == "v1"sv) {
+        config.version = NGTCP2_PROTO_VER_V1;
+        break;
+      }
+      if (optarg == "v2draft"sv) {
+        config.version = NGTCP2_PROTO_VER_V2_DRAFT;
+        break;
+      }
+      auto rv = util::parse_version(optarg);
+      if (!rv) {
+        std::cerr << "version: invalid version " << std::quoted(optarg)
+                  << std::endl;
+        exit(EXIT_FAILURE);
+      }
+      config.version = *rv;
       break;
+    }
     case '?':
       print_usage();
       exit(EXIT_FAILURE);
@@ -2758,6 +2837,83 @@ int main(int argc, char **argv) {
           config.handshake_timeout = *t;
         }
         break;
+      case 37: {
+        // --other-versions
+        if (strlen(optarg) == 0) {
+          config.other_versions.resize(0);
+          break;
+        }
+        auto l = util::split_str(optarg);
+        config.other_versions.resize(l.size());
+        auto it = std::begin(config.other_versions);
+        for (const auto &k : l) {
+          if (k == "v1"sv) {
+            *it++ = NGTCP2_PROTO_VER_V1;
+            continue;
+          }
+          if (k == "v2draft"sv) {
+            *it++ = NGTCP2_PROTO_VER_V2_DRAFT;
+            continue;
+          }
+          auto rv = util::parse_version(k);
+          if (!rv) {
+            std::cerr << "other-versions: invalid version " << std::quoted(k)
+                      << std::endl;
+            exit(EXIT_FAILURE);
+          }
+          *it++ = *rv;
+        }
+        break;
+      }
+      case 38:
+        // --no-pmtud
+        config.no_pmtud = true;
+        break;
+      case 39: {
+        // --preferred-versions
+        auto l = util::split_str(optarg);
+        if (l.size() > max_preferred_versionslen) {
+          std::cerr << "preferred-versions: too many versions > "
+                    << max_preferred_versionslen << std::endl;
+        }
+        config.preferred_versions.resize(l.size());
+        auto it = std::begin(config.preferred_versions);
+        for (const auto &k : l) {
+          if (k == "v1"sv) {
+            *it++ = NGTCP2_PROTO_VER_V1;
+            continue;
+          }
+          if (k == "v2draft"sv) {
+            *it++ = NGTCP2_PROTO_VER_V2_DRAFT;
+            continue;
+          }
+          auto rv = util::parse_version(k);
+          if (!rv) {
+            std::cerr << "preferred-versions: invalid version "
+                      << std::quoted(k) << std::endl;
+            exit(EXIT_FAILURE);
+          }
+          if (!ngtcp2_is_supported_version(*rv)) {
+            std::cerr << "preferred-versions: unsupported version "
+                      << std::quoted(k) << std::endl;
+            exit(EXIT_FAILURE);
+          }
+          *it++ = *rv;
+        }
+        break;
+      }
+      case 40:
+        // --ack-thresh
+        if (auto n = util::parse_uint(optarg); !n) {
+          std::cerr << "ack-thresh: invalid argument" << std::endl;
+          exit(EXIT_FAILURE);
+        } else if (*n > 100) {
+          std::cerr << "ack-thresh: must not exceed 100" << std::endl;
+          exit(EXIT_FAILURE);
+        } else {
+          config.ack_thresh = *n;
+        }
+        break;
       }
       break;
     default:
@@ -2814,6 +2970,26 @@ int main(int argc, char **argv) {
     exit(EXIT_FAILURE);
   }
 
+  if (!ngtcp2_is_reserved_version(config.version)) {
+    if (!config.preferred_versions.empty() &&
+        std::find(std::begin(config.preferred_versions),
+                  std::end(config.preferred_versions),
+                  config.version) == std::end(config.preferred_versions)) {
+      std::cerr << "preferred-version: must include version "
+                << "0x" << config.version << std::endl;
+      exit(EXIT_FAILURE);
+    }
+
+    if (!config.other_versions.empty() &&
+        std::find(std::begin(config.other_versions),
+                  std::end(config.other_versions),
+                  config.version) == std::end(config.other_versions)) {
+      std::cerr << "other-versions: must include version "
+                << "0x" << config.version << std::endl;
+      exit(EXIT_FAILURE);
+    }
+  }
+
   if (config.nstreams == 0) {
     config.nstreams = config.requests.size();
   }
@@ -2839,10 +3015,37 @@ int main(int argc, char **argv) {
     exit(EXIT_FAILURE);
   }
 
-  Client c(EV_DEFAULT);
+  auto client_chosen_version = config.version;
 
-  if (run(c, addr, port, tls_ctx) != 0) {
-    exit(EXIT_FAILURE);
+  for (;;) {
+    Client c(EV_DEFAULT, client_chosen_version, config.version);
+
+    if (run(c, addr, port, tls_ctx) != 0) {
+      exit(EXIT_FAILURE);
+    }
+
+    if (config.preferred_versions.empty()) {
+      break;
+    }
+
+    auto &offered_versions = c.get_offered_versions();
+    if (offered_versions.empty()) {
+      break;
+    }
+
+    client_chosen_version = ngtcp2_select_version(
+        config.preferred_versions.data(), config.preferred_versions.size(),
+        offered_versions.data(), offered_versions.size());
+
+    if (client_chosen_version == 0) {
+      std::cerr << "Unable to select a version" << std::endl;
+      exit(EXIT_FAILURE);
+    }
+
+    if (!config.quiet) {
+      std::cerr << "Client selected version " << std::hex << "0x"
+                << client_chosen_version << std::dec << std::endl;
+    }
   }
 
   return EXIT_SUCCESS;
