@@ -72,12 +72,12 @@ struct Endpoint {
 
 class Client : public ClientBase {
 public:
-  Client(struct ev_loop *loop);
+  Client(struct ev_loop *loop, uint32_t client_chosen_version,
+         uint32_t original_version);
   ~Client();
 
   int init(int fd, const Address &local_addr, const Address &remote_addr,
-           const char *addr, const char *port, uint32_t version,
-           TLSClientContext &tls_ctx);
+           const char *addr, const char *port, TLSClientContext &tls_ctx);
   void disconnect();
 
   int on_read(const Endpoint &ep);
@@ -86,9 +86,10 @@ public:
   int feed_data(const Endpoint &ep, const sockaddr *sa, socklen_t salen,
                 const ngtcp2_pkt_info *pi, uint8_t *data, size_t datalen);
   int handle_expiry();
-  void schedule_retransmit();
+  void update_timer();
   int handshake_completed();
   int handshake_confirmed();
+  void recv_version_negotiation(const uint32_t *sv, size_t nsv);
 
   int send_packet(const Endpoint &ep, const ngtcp2_addr &remote_addr,
                   unsigned int ecn, const uint8_t *data, size_t datalen);
@@ -125,29 +126,31 @@ public:
   int on_stream_stop_sending(int64_t stream_id);
   int extend_max_stream_data(int64_t stream_id, uint64_t max_data);
   int stop_sending(int64_t stream_id, uint64_t app_error_code);
+  int reset_stream(int64_t stream_id, uint64_t app_error_code);
   int http_stream_close(int64_t stream_id, uint64_t app_error_code);
 
-  void reset_idle_timer();
-
-  void idle_timeout();
-
   void on_send_blocked(const Endpoint &ep, const ngtcp2_addr &remote_addr,
                        unsigned int ecn, size_t datalen);
   void start_wev_endpoint(const Endpoint &ep);
   int send_blocked_packet();
 
+  const std::vector<uint32_t> &get_offered_versions() const;
+
+  bool get_early_data() const;
+  void early_data_rejected();
+
 private:
   std::vector<Endpoint> endpoints_;
   Address remote_addr_;
   ev_io wev_;
   ev_timer timer_;
-  ev_timer rttimer_;
   ev_timer change_local_addr_timer_;
   ev_timer key_update_timer_;
   ev_timer delay_stream_timer_;
   ev_signal sigintev_;
   struct ev_loop *loop_;
   std::map<int64_t, std::unique_ptr<Stream>> streams_;
+  std::vector<uint32_t> offered_versions_;
   nghttp3_conn *httpconn_;
   // addr_ is the server host address.
   const char *addr_;
@@ -159,12 +162,16 @@ private:
   size_t nstreams_closed_;
   // nkey_update_ is the number of key update occurred.
   size_t nkey_update_;
-  uint32_t version_;
+  uint32_t client_chosen_version_;
+  uint32_t original_version_;
   // early_data_ is true if client attempts to do 0RTT data transfer.
   bool early_data_;
   // should_exit_ is true if client should exit rather than waiting
   // for timeout.
   bool should_exit_;
+  // should_exit_on_handshake_confirmed_ is true if client should exit
+  // when handshake confirmed.
+  bool should_exit_on_handshake_confirmed_;
   // handshake_confirmed_ gets true after handshake has been
   // confirmed.
   bool handshake_confirmed_;
