@@ -29,20 +29,24 @@
 #include <iostream>
 #include <fstream>
 
-#include <ngtcp2/ngtcp2_crypto.h>
-
 #include "debug.h"
 #include "template.h"
 #include "util.h"
 
 using namespace ngtcp2;
+using namespace std::literals;
 
 extern Config config;
 
+static ngtcp2_conn *get_conn(ngtcp2_crypto_conn_ref *conn_ref) {
+  auto c = static_cast<ClientBase *>(conn_ref->user_data);
+  return c->conn();
+}
+
 ClientBase::ClientBase()
-    : qlog_(nullptr),
-      conn_(nullptr),
-      last_error_{QUICErrorType::Transport, 0} {}
+    : conn_ref_{get_conn, this}, qlog_(nullptr), conn_(nullptr) {
+  ngtcp2_connection_close_error_default(&last_error_);
+}
 
 ClientBase::~ClientBase() {
   if (conn_) {
@@ -90,9 +94,9 @@ int ClientBase::read_transport_params(const char *path,
   }
 
   for (std::string line; std::getline(f, line);) {
-    if (util::istarts_with_l(line, "initial_max_streams_bidi=")) {
+    if (util::istarts_with(line, "initial_max_streams_bidi="sv)) {
       if (auto n = util::parse_uint(line.c_str() +
-                                    str_size("initial_max_streams_bidi="));
+                                    "initial_max_streams_bidi="sv.size());
           !n) {
         return -1;
       } else {
@@ -101,9 +105,9 @@ int ClientBase::read_transport_params(const char *path,
       continue;
     }
 
-    if (util::istarts_with_l(line, "initial_max_streams_uni=")) {
+    if (util::istarts_with(line, "initial_max_streams_uni="sv)) {
       if (auto n = util::parse_uint(line.c_str() +
-                                    str_size("initial_max_streams_uni="));
+                                    "initial_max_streams_uni="sv.size());
           !n) {
         return -1;
       } else {
@@ -112,9 +116,9 @@ int ClientBase::read_transport_params(const char *path,
       continue;
     }
 
-    if (util::istarts_with_l(line, "initial_max_stream_data_bidi_local=")) {
+    if (util::istarts_with(line, "initial_max_stream_data_bidi_local="sv)) {
       if (auto n = util::parse_uint(
-              line.c_str() + str_size("initial_max_stream_data_bidi_local="));
+              line.c_str() + "initial_max_stream_data_bidi_local="sv.size());
           !n) {
         return -1;
       } else {
@@ -123,9 +127,9 @@ int ClientBase::read_transport_params(const char *path,
       continue;
     }
 
-    if (util::istarts_with_l(line, "initial_max_stream_data_bidi_remote=")) {
+    if (util::istarts_with(line, "initial_max_stream_data_bidi_remote="sv)) {
       if (auto n = util::parse_uint(
-              line.c_str() + str_size("initial_max_stream_data_bidi_remote="));
+              line.c_str() + "initial_max_stream_data_bidi_remote="sv.size());
           !n) {
         return -1;
       } else {
@@ -134,9 +138,9 @@ int ClientBase::read_transport_params(const char *path,
       continue;
     }
 
-    if (util::istarts_with_l(line, "initial_max_stream_data_uni=")) {
+    if (util::istarts_with(line, "initial_max_stream_data_uni="sv)) {
       if (auto n = util::parse_uint(line.c_str() +
-                                    str_size("initial_max_stream_data_uni="));
+                                    "initial_max_stream_data_uni="sv.size());
           !n) {
         return -1;
       } else {
@@ -145,9 +149,9 @@ int ClientBase::read_transport_params(const char *path,
       continue;
     }
 
-    if (util::istarts_with_l(line, "initial_max_data=")) {
+    if (util::istarts_with(line, "initial_max_data="sv)) {
       if (auto n =
-              util::parse_uint(line.c_str() + str_size("initial_max_data="));
+              util::parse_uint(line.c_str() + "initial_max_data="sv.size());
           !n) {
         return -1;
       } else {
@@ -156,9 +160,9 @@ int ClientBase::read_transport_params(const char *path,
       continue;
     }
 
-    if (util::istarts_with_l(line, "active_connection_id_limit=")) {
+    if (util::istarts_with(line, "active_connection_id_limit="sv)) {
       if (auto n = util::parse_uint(line.c_str() +
-                                    str_size("active_connection_id_limit="));
+                                    "active_connection_id_limit="sv.size());
           !n) {
         return -1;
       } else {
@@ -167,9 +171,9 @@ int ClientBase::read_transport_params(const char *path,
       continue;
     }
 
-    if (util::istarts_with_l(line, "max_datagram_frame_size=")) {
+    if (util::istarts_with(line, "max_datagram_frame_size="sv)) {
       if (auto n = util::parse_uint(line.c_str() +
-                                    str_size("max_datagram_frame_size="));
+                                    "max_datagram_frame_size="sv.size());
           !n) {
         return -1;
       } else {
@@ -182,95 +186,6 @@ int ClientBase::read_transport_params(const char *path,
   return 0;
 }
 
-int ClientBase::on_rx_key(ngtcp2_crypto_level level, const uint8_t *secret,
-                          size_t secretlen) {
-  std::array<uint8_t, 64> key, iv, hp_key;
-
-  if (ngtcp2_crypto_derive_and_install_rx_key(conn_, key.data(), iv.data(),
-                                              hp_key.data(), level, secret,
-                                              secretlen) != 0) {
-    return -1;
-  }
-
-  auto crypto_ctx = level == NGTCP2_CRYPTO_LEVEL_EARLY
-                        ? ngtcp2_conn_get_early_crypto_ctx(conn_)
-                        : ngtcp2_conn_get_crypto_ctx(conn_);
-  auto aead = &crypto_ctx->aead;
-  auto keylen = ngtcp2_crypto_aead_keylen(aead);
-  auto ivlen = ngtcp2_crypto_packet_protection_ivlen(aead);
-
-  const char *title = nullptr;
-  switch (level) {
-  case NGTCP2_CRYPTO_LEVEL_EARLY:
-    assert(0);
-  case NGTCP2_CRYPTO_LEVEL_HANDSHAKE:
-    title = "handshake_traffic";
-    break;
-  case NGTCP2_CRYPTO_LEVEL_APPLICATION:
-    title = "application_traffic";
-    break;
-  default:
-    assert(0);
-  }
-
-  if (!config.quiet && config.show_secret) {
-    std::cerr << title << " rx secret" << std::endl;
-    debug::print_secrets(secret, secretlen, key.data(), keylen, iv.data(),
-                         ivlen, hp_key.data(), keylen);
-  }
-
-  if (level == NGTCP2_CRYPTO_LEVEL_APPLICATION) {
-    if (config.tp_file) {
-      ngtcp2_transport_params params;
-
-      ngtcp2_conn_get_remote_transport_params(conn_, &params);
-
-      if (write_transport_params(config.tp_file, &params) != 0) {
-        std::cerr << "Could not write transport parameters in "
-                  << config.tp_file << std::endl;
-      }
-    }
-  }
-
-  return 0;
-}
-
-int ClientBase::on_tx_key(ngtcp2_crypto_level level, const uint8_t *secret,
-                          size_t secretlen) {
-  std::array<uint8_t, 64> key, iv, hp_key;
-
-  if (ngtcp2_crypto_derive_and_install_tx_key(conn_, key.data(), iv.data(),
-                                              hp_key.data(), level, secret,
-                                              secretlen) != 0) {
-    return -1;
-  }
-
-  auto crypto_ctx = level == NGTCP2_CRYPTO_LEVEL_EARLY
-                        ? ngtcp2_conn_get_early_crypto_ctx(conn_)
-                        : ngtcp2_conn_get_crypto_ctx(conn_);
-  auto aead = &crypto_ctx->aead;
-  auto keylen = ngtcp2_crypto_aead_keylen(aead);
-  auto ivlen = ngtcp2_crypto_packet_protection_ivlen(aead);
-  auto title = debug::secret_title(level);
-
-  if (!config.quiet && config.show_secret) {
-    std::cerr << title << " tx secret" << std::endl;
-    debug::print_secrets(secret, secretlen, key.data(), keylen, iv.data(),
-                         ivlen, hp_key.data(), keylen);
-  }
-
-  return 0;
-}
-
-void ClientBase::write_client_handshake(ngtcp2_crypto_level level,
-                                        const uint8_t *data, size_t datalen) {
-  ngtcp2_conn_submit_crypto_data(conn_, level, data, datalen);
-}
-
-void ClientBase::set_tls_alert(uint8_t alert) {
-  last_error_ = quic_err_tls(alert);
-}
-
 ngtcp2_conn *ClientBase::conn() const { return conn_; }
 
 void qlog_write_cb(void *user_data, uint32_t flags, const void *data,
@@ -284,15 +199,4 @@ void ClientBase::write_qlog(const void *data, size_t datalen) {
   fwrite(data, 1, datalen, qlog_);
 }
 
-int ClientBase::call_application_rx_key_cb() const {
-  if (!application_rx_key_cb_) {
-    return 0;
-  }
-  return application_rx_key_cb_();
-}
-
-void ClientBase::process_unhandled_tls_alert() {
-  if (auto alert = tls_session_.get_tls_alert(); alert) {
-    set_tls_alert(alert);
-  }
-}
+ngtcp2_crypto_conn_ref *ClientBase::conn_ref() { return &conn_ref_; }
