@@ -35,7 +35,7 @@
 #include <string_view>
 #include <functional>
 
-#include <ngtcp2/ngtcp2.h>
+#include <ngtcp2/ngtcp2_crypto.h>
 
 #include "tls_client_session.h"
 #include "network.h"
@@ -169,6 +169,18 @@ struct Config {
   // handshake_timeout is the period of time before giving up QUIC
   // connection establishment.
   ngtcp2_duration handshake_timeout;
+  // preferred_versions includes QUIC versions in the order of
+  // preference.  Client uses this field to select a version from the
+  // version set offered in Version Negotiation packet.
+  std::vector<uint32_t> preferred_versions;
+  // other_versions includes QUIC versions that are sent in
+  // other_versions field of version_information transport_parameter.
+  std::vector<uint32_t> other_versions;
+  // no_pmtud disables Path MTU Discovery.
+  bool no_pmtud;
+  // ack_thresh is the minimum number of the received ACK eliciting
+  // packets that triggers immediate acknowledgement.
+  size_t ack_thresh;
 };
 
 class ClientBase {
@@ -178,30 +190,20 @@ public:
 
   ngtcp2_conn *conn() const;
 
-  int on_rx_key(ngtcp2_crypto_level level, const uint8_t *secret,
-                size_t secretlen);
-  int on_tx_key(ngtcp2_crypto_level level, const uint8_t *secret,
-                size_t secretlen);
-  void write_client_handshake(ngtcp2_crypto_level crypto_level,
-                              const uint8_t *data, size_t datalen);
-  void set_tls_alert(uint8_t alert);
-
   int write_transport_params(const char *path,
                              const ngtcp2_transport_params *params);
   int read_transport_params(const char *path, ngtcp2_transport_params *params);
 
   void write_qlog(const void *data, size_t datalen);
 
-  int call_application_rx_key_cb() const;
-
-  void process_unhandled_tls_alert();
+  ngtcp2_crypto_conn_ref *conn_ref();
 
 protected:
+  ngtcp2_crypto_conn_ref conn_ref_;
   TLSClientSession tls_session_;
   FILE *qlog_;
   ngtcp2_conn *conn_;
-  QUICError last_error_;
-  std::function<int()> application_rx_key_cb_;
+  ngtcp2_connection_close_error last_error_;
 };
 
 void qlog_write_cb(void *user_data, uint32_t flags, const void *data,
