@@ -116,7 +116,7 @@ public:
   int feed_data(const Endpoint &ep, const Address &local_addr,
                 const sockaddr *sa, socklen_t salen, const ngtcp2_pkt_info *pi,
                 uint8_t *data, size_t datalen);
-  void schedule_retransmit();
+  void update_timer();
   int handle_expiry();
   void signal_write();
   int handshake_completed();
@@ -131,7 +131,6 @@ public:
   int on_stream_close(int64_t stream_id, uint64_t app_error_code);
   void start_draining_period();
   int start_closing_period();
-  bool draining() const;
   int handle_error();
   int send_conn_close();
 
@@ -145,15 +144,12 @@ public:
   int extend_max_stream_data(int64_t stream_id, uint64_t max_data);
   void shutdown_read(int64_t stream_id, int app_error_code);
 
-  void reset_idle_timer();
-
   void write_qlog(const void *data, size_t datalen);
   void add_sendq(Stream *stream);
 
   void on_send_blocked(Endpoint &ep, const ngtcp2_addr &local_addr,
                        const ngtcp2_addr &remote_addr, unsigned int ecn,
-                       const uint8_t *data, size_t datalen,
-                       size_t max_udp_payload_size);
+                       const uint8_t *data, size_t datalen, size_t gso_size);
   void start_wev_endpoint(const Endpoint &ep);
   int send_blocked_packet();
 
@@ -162,7 +158,6 @@ private:
   Server *server_;
   ev_io wev_;
   ev_timer timer_;
-  ev_timer rttimer_;
   FILE *qlog_;
   ngtcp2_cid scid_;
   std::unordered_map<int64_t, std::unique_ptr<Stream>> streams_;
@@ -173,8 +168,7 @@ private:
   std::unique_ptr<Buffer> conn_closebuf_;
   // nkey_update_ is the number of key update occurred.
   size_t nkey_update_;
-  // draining_ becomes true when draining period starts.
-  bool draining_;
+  bool no_gso_;
 
   struct {
     bool send_blocked;
@@ -188,7 +182,7 @@ private:
       unsigned int ecn;
       const uint8_t *data;
       size_t datalen;
-      size_t max_udp_payload_size;
+      size_t gso_size;
     } blocked[2];
     std::unique_ptr<uint8_t[]> data;
   } tx_;
@@ -210,8 +204,8 @@ public:
                                const Address &local_addr, const sockaddr *sa,
                                socklen_t salen);
   int send_retry(const ngtcp2_pkt_hd *chd, Endpoint &ep,
-                 const Address &local_addr, const sockaddr *sa,
-                 socklen_t salen);
+                 const Address &local_addr, const sockaddr *sa, socklen_t salen,
+                 size_t max_pktlen);
   int send_stateless_connection_close(const ngtcp2_pkt_hd *chd, Endpoint &ep,
                                       const Address &local_addr,
                                       const sockaddr *sa, socklen_t salen);
@@ -221,7 +215,12 @@ public:
                    socklen_t salen);
   int send_packet(Endpoint &ep, const ngtcp2_addr &local_addr,
                   const ngtcp2_addr &remote_addr, unsigned int ecn,
-                  const uint8_t *data, size_t datalen, size_t gso_size);
+                  const uint8_t *data, size_t datalen);
+  std::pair<size_t, int> send_packet(Endpoint &ep, bool &no_gso,
+                                     const ngtcp2_addr &local_addr,
+                                     const ngtcp2_addr &remote_addr,
+                                     unsigned int ecn, const uint8_t *data,
+                                     size_t datalen, size_t gso_size);
   void remove(const Handler *h);
 
   void associate_cid(const ngtcp2_cid *cid, Handler *h);
