@@ -122,7 +122,7 @@ public:
   int feed_data(const Endpoint &ep, const Address &local_addr,
                 const sockaddr *sa, socklen_t salen, const ngtcp2_pkt_info *pi,
                 uint8_t *data, size_t datalen);
-  void schedule_retransmit();
+  void update_timer();
   int handle_expiry();
   void signal_write();
   int handshake_completed();
@@ -136,7 +136,6 @@ public:
   int on_stream_close(int64_t stream_id, uint64_t app_error_code);
   void start_draining_period();
   int start_closing_period();
-  bool draining() const;
   int handle_error();
   int send_conn_close();
 
@@ -165,14 +164,11 @@ public:
   int http_stop_sending(int64_t stream_id, uint64_t app_error_code);
   int http_reset_stream(int64_t stream_id, uint64_t app_error_code);
 
-  void reset_idle_timer();
-
   void write_qlog(const void *data, size_t datalen);
 
   void on_send_blocked(Endpoint &ep, const ngtcp2_addr &local_addr,
                        const ngtcp2_addr &remote_addr, unsigned int ecn,
-                       const uint8_t *data, size_t datalen,
-                       size_t max_udp_payload_size);
+                       const uint8_t *data, size_t datalen, size_t gso_size);
   void start_wev_endpoint(const Endpoint &ep);
   int send_blocked_packet();
 
@@ -181,7 +177,6 @@ private:
   Server *server_;
   ev_io wev_;
   ev_timer timer_;
-  ev_timer rttimer_;
   FILE *qlog_;
   ngtcp2_cid scid_;
   nghttp3_conn *httpconn_;
@@ -192,8 +187,7 @@ private:
   std::unique_ptr<Buffer> conn_closebuf_;
   // nkey_update_ is the number of key update occurred.
   size_t nkey_update_;
-  // draining_ becomes true when draining period starts.
-  bool draining_;
+  bool no_gso_;
 
   struct {
     bool send_blocked;
@@ -207,7 +201,7 @@ private:
       unsigned int ecn;
       const uint8_t *data;
       size_t datalen;
-      size_t max_udp_payload_size;
+      size_t gso_size;
     } blocked[2];
     std::unique_ptr<uint8_t[]> data;
   } tx_;
@@ -229,8 +223,8 @@ public:
                                const Address &local_addr, const sockaddr *sa,
                                socklen_t salen);
   int send_retry(const ngtcp2_pkt_hd *chd, Endpoint &ep,
-                 const Address &local_addr, const sockaddr *sa,
-                 socklen_t salen);
+                 const Address &local_addr, const sockaddr *sa, socklen_t salen,
+                 size_t max_pktlen);
   int send_stateless_connection_close(const ngtcp2_pkt_hd *chd, Endpoint &ep,
                                       const Address &local_addr,
                                       const sockaddr *sa, socklen_t salen);
@@ -240,7 +234,12 @@ public:
                    socklen_t salen);
   int send_packet(Endpoint &ep, const ngtcp2_addr &local_addr,
                   const ngtcp2_addr &remote_addr, unsigned int ecn,
-                  const uint8_t *data, size_t datalen, size_t gso_size);
+                  const uint8_t *data, size_t datalen);
+  std::pair<size_t, int> send_packet(Endpoint &ep, bool &no_gso,
+                                     const ngtcp2_addr &local_addr,
+                                     const ngtcp2_addr &remote_addr,
+                                     unsigned int ecn, const uint8_t *data,
+                                     size_t datalen, size_t gso_size);
   void remove(const Handler *h);
 
   void associate_cid(const ngtcp2_cid *cid, Handler *h);
