@@ -132,6 +132,7 @@ static int connect_sock(struct sockaddr *local_addr, socklen_t *plocal_addrlen,
 }
 
 struct client {
+  ngtcp2_crypto_conn_ref conn_ref;
   int fd;
   struct sockaddr_storage local_addr;
   socklen_t local_addrlen;
@@ -146,73 +147,10 @@ struct client {
     size_t nwrite;
   } stream;
 
-  uint64_t last_error;
+  ngtcp2_connection_close_error last_error;
 
   ev_io rev;
   ev_timer timer;
-  ev_timer idle_timer;
-};
-
-static int set_encryption_secrets(SSL *ssl, OSSL_ENCRYPTION_LEVEL ossl_level,
-                                  const uint8_t *rx_secret,
-                                  const uint8_t *tx_secret, size_t secretlen) {
-  struct client *c = SSL_get_app_data(ssl);
-  ngtcp2_crypto_level level =
-      ngtcp2_crypto_openssl_from_ossl_encryption_level(ossl_level);
-
-  if (rx_secret &&
-      ngtcp2_crypto_derive_and_install_rx_key(c->conn, NULL, NULL, NULL, level,
-                                              rx_secret, secretlen) != 0) {
-    fprintf(stderr, "ngtcp2_crypto_derive_and_install_rx_key failed\n");
-    return 0;
-  }
-
-  if (ngtcp2_crypto_derive_and_install_tx_key(c->conn, NULL, NULL, NULL, level,
-                                              tx_secret, secretlen) != 0) {
-    fprintf(stderr, "ngtcp2_crypto_derive_and_install_tx_key failed\n");
-    return 0;
-  }
-
-  return 1;
-}
-
-static int add_handshake_data(SSL *ssl, OSSL_ENCRYPTION_LEVEL ossl_level,
-                              const uint8_t *data, size_t len) {
-  struct client *c = SSL_get_app_data(ssl);
-  ngtcp2_crypto_level level =
-      ngtcp2_crypto_openssl_from_ossl_encryption_level(ossl_level);
-  int rv;
-
-  rv = ngtcp2_conn_submit_crypto_data(c->conn, level, data, len);
-  if (rv != 0) {
-    fprintf(stderr, "ngtcp2_conn_submit_crypto_data: %s\n",
-            ngtcp2_strerror(rv));
-    return 0;
-  }
-
-  return 1;
-}
-
-static int flush_flight(SSL *ssl) {
-  (void)ssl;
-  return 1;
-}
-
-static int send_alert(SSL *ssl, OSSL_ENCRYPTION_LEVEL ossl_level,
-                      uint8_t alert) {
-  struct client *c = SSL_get_app_data(ssl);
-  (void)ossl_level;
-
-  c->last_error = NGTCP2_CRYPTO_ERROR | alert;
-
-  return 1;
-}
-
-static SSL_QUIC_METHOD quic_method = {
-    set_encryption_secrets,
-    add_handshake_data,
-    flush_flight,
-    send_alert,
 };
 
 static int numeric_host_family(const char *hostname, int family) {
@@ -233,9 +171,10 @@ static int client_ssl_init(struct client *c) {
     return -1;
   }
 
-  SSL_CTX_set_min_proto_version(c->ssl_ctx, TLS1_3_VERSION);
-  SSL_CTX_set_max_proto_version(c->ssl_ctx, TLS1_3_VERSION);
-  SSL_CTX_set_quic_method(c->ssl_ctx, &quic_method);
+  if (ngtcp2_crypto_openssl_configure_client_context(c->ssl_ctx) != 0) {
+    fprintf(stderr, "ngtcp2_crypto_openssl_configure_client_context failed\n");
+    return -1;
+  }
 
   c->ssl = SSL_new(c->ssl_ctx);
   if (!c->ssl) {
@@ -243,13 +182,16 @@ static int client_ssl_init(struct client *c) {
     return -1;
   }
 
-  SSL_set_app_data(c->ssl, c);
+  SSL_set_app_data(c->ssl, &c->conn_ref);
   SSL_set_connect_state(c->ssl);
   SSL_set_alpn_protos(c->ssl, (const unsigned char *)ALPN, sizeof(ALPN) - 1);
   if (!numeric_host(REMOTE_HOST)) {
     SSL_set_tlsext_host_name(c->ssl, REMOTE_HOST);
   }
 
+  /* For NGTCP2_PROTO_VER_V1 */
+  SSL_set_quic_transport_version(c->ssl, TLSEXT_TYPE_quic_transport_parameters);
+
   return 0;
 }
 
@@ -378,6 +320,10 @@ static int client_quic_init(struct client *c,
       NULL, /* lost_datagram */
       ngtcp2_crypto_get_path_challenge_data_cb,
       NULL, /* stream_stop_sending */
+      ngtcp2_crypto_version_negotiation_cb,
+      NULL, /* recv_rx_key */
+      NULL, /* recv_tx_key */
+      NULL, /* early_data_rejected */
   };
   ngtcp2_cid dcid, scid;
   ngtcp2_settings settings;
@@ -420,16 +366,6 @@ static int client_quic_init(struct client *c,
   return 0;
 }
 
-static void client_reset_idle_timer(struct client *c) {
-  ngtcp2_tstamp expiry = ngtcp2_conn_get_idle_expiry(c->conn);
-  ngtcp2_tstamp now = timestamp();
-
-  c->idle_timer.repeat =
-      expiry < now ? 1e-9 : (ev_tstamp)(expiry - now) / NGTCP2_SECONDS;
-
-  ev_timer_again(EV_DEFAULT, &c->idle_timer);
-}
-
 static int client_read(struct client *c) {
   uint8_t buf[65536];
   struct sockaddr_storage addr;
@@ -466,25 +402,19 @@ static int client_read(struct client *c) {
                               timestamp());
     if (rv != 0) {
       fprintf(stderr, "ngtcp2_conn_read_pkt: %s\n", ngtcp2_strerror(rv));
-      switch (rv) {
-      case NGTCP2_ERR_REQUIRED_TRANSPORT_PARAM:
-      case NGTCP2_ERR_MALFORMED_TRANSPORT_PARAM:
-      case NGTCP2_ERR_TRANSPORT_PARAM:
-      case NGTCP2_ERR_PROTO:
-        c->last_error = ngtcp2_err_infer_quic_transport_error_code(rv);
-        break;
-      default:
-        if (!c->last_error) {
-          c->last_error = ngtcp2_err_infer_quic_transport_error_code(rv);
+      if (!c->last_error.error_code) {
+        if (rv == NGTCP2_ERR_CRYPTO) {
+          ngtcp2_connection_close_error_set_transport_error_tls_alert(
+              &c->last_error, ngtcp2_conn_get_tls_alert(c->conn), NULL, 0);
+        } else {
+          ngtcp2_connection_close_error_set_transport_error_liberr(
+              &c->last_error, rv, NULL, 0);
         }
-        break;
       }
       return -1;
     }
   }
 
-  client_reset_idle_timer(c);
-
   return 0;
 }
 
@@ -567,7 +497,8 @@ static int client_write_streams(struct client *c) {
       default:
         fprintf(stderr, "ngtcp2_conn_writev_stream: %s\n",
                 ngtcp2_strerror((int)nwrite));
-        c->last_error = ngtcp2_err_infer_quic_transport_error_code((int)nwrite);
+        ngtcp2_connection_close_error_set_transport_error_liberr(
+            &c->last_error, (int)nwrite, NULL, 0);
         return -1;
       }
     }
@@ -585,8 +516,6 @@ static int client_write_streams(struct client *c) {
     }
   }
 
-  client_reset_idle_timer(c);
-
   return 0;
 }
 
@@ -625,15 +554,15 @@ static void client_close(struct client *c) {
   ngtcp2_path_storage ps;
   uint8_t buf[1280];
 
-  if (ngtcp2_conn_is_in_closing_period(c->conn) || !c->last_error) {
+  if (ngtcp2_conn_is_in_closing_period(c->conn) ||
+      ngtcp2_conn_is_in_draining_period(c->conn)) {
     goto fin;
   }
 
   ngtcp2_path_storage_zero(&ps);
 
-  nwrite = ngtcp2_conn_write_connection_close(c->conn, &ps.path, &pi, buf,
-                                              sizeof(buf), c->last_error, NULL,
-                                              0, timestamp());
+  nwrite = ngtcp2_conn_write_connection_close(
+      c->conn, &ps.path, &pi, buf, sizeof(buf), &c->last_error, timestamp());
   if (nwrite < 0) {
     fprintf(stderr, "ngtcp2_conn_write_connection_close: %s\n",
             ngtcp2_strerror((int)nwrite));
@@ -676,14 +605,9 @@ static void timer_cb(struct ev_loop *loop, ev_timer *w, int revents) {
   }
 }
 
-static void idle_timer_cb(struct ev_loop *loop, ev_timer *w, int revents) {
-  (void)loop;
-  (void)w;
-  (void)revents;
-
-  fprintf(stderr, "idle timeout\n");
-
-  ev_break(EV_DEFAULT, EVBREAK_ALL);
+static ngtcp2_conn *get_conn(ngtcp2_crypto_conn_ref *conn_ref) {
+  struct client *c = conn_ref->user_data;
+  return c->conn;
 }
 
 static int client_init(struct client *c) {
@@ -692,6 +616,8 @@ static int client_init(struct client *c) {
 
   memset(c, 0, sizeof(*c));
 
+  ngtcp2_connection_close_error_default(&c->last_error);
+
   c->fd = create_sock((struct sockaddr *)&remote_addr, &remote_addrlen,
                       REMOTE_HOST, REMOTE_PORT);
   if (c->fd == -1) {
@@ -717,6 +643,9 @@ static int client_init(struct client *c) {
 
   c->stream.stream_id = -1;
 
+  c->conn_ref.get_conn = get_conn;
+  c->conn_ref.user_data = c;
+
   ev_io_init(&c->rev, read_cb, c->fd, EV_READ);
   c->rev.data = c;
   ev_io_start(EV_DEFAULT, &c->rev);
@@ -724,10 +653,6 @@ static int client_init(struct client *c) {
   ev_timer_init(&c->timer, timer_cb, 0., 0.);
   c->timer.data = c;
 
-  ev_timer_init(&c->idle_timer, idle_timer_cb, 0., 30.);
-  c->idle_timer.data = c;
-  ev_timer_again(EV_DEFAULT, &c->timer);
-
   return 0;
 }
 
