@@ -45,4 +45,8 @@
 #  include "tls_client_context_picotls.h"
 #endif // ENABLE_EXAMPLE_PICOTLS && WITH_EXAMPLE_PICOTLS
 
-#endif // TLS_CLIENT_CONTEXT_OPENSSL_H
+#if defined(ENABLE_EXAMPLE_WOLFSSL) && defined(WITH_EXAMPLE_WOLFSSL)
+#  include "tls_client_context_wolfssl.h"
+#endif // ENABLE_EXAMPLE_WOLFSSL && WITH_EXAMPLE_WOLFSSL
+
+#endif // TLS_CLIENT_CONTEXT_H
