@@ -45,4 +45,8 @@
 #  include "tls_server_context_picotls.h"
 #endif // ENABLE_EXAMPLE_PICOTLS && WITH_EXAMPLE_PICOTLS
 
+#if defined(ENABLE_EXAMPLE_WOLFSSL) && defined(WITH_EXAMPLE_WOLFSSL)
+#  include "tls_server_context_wolfssl.h"
+#endif // ENABLE_EXAMPLE_WOLFSSL && WITH_EXAMPLE_WOLFSSL
+
 #endif // TLS_SERVER_CONTEXT_H
