@@ -47,35 +47,44 @@ namespace ngtcp2 {
 
 namespace util {
 
-template <typename T, size_t N1, size_t N2>
-constexpr nghttp3_nv make_nv(const T (&name)[N1], const T (&value)[N2]) {
-  return nghttp3_nv{(uint8_t *)name, (uint8_t *)value, N1 - 1, N2 - 1,
-                    NGHTTP3_NV_FLAG_NONE};
+inline nghttp3_nv make_nv(const std::string_view &name,
+                          const std::string_view &value, uint8_t flags) {
+  return nghttp3_nv{
+      reinterpret_cast<uint8_t *>(const_cast<char *>(std::data(name))),
+      reinterpret_cast<uint8_t *>(const_cast<char *>(std::data(value))),
+      name.size(),
+      value.size(),
+      flags,
+  };
 }
 
-template <typename T, size_t N, typename S>
-constexpr nghttp3_nv make_nv(const T (&name)[N], const S &value) {
-  return nghttp3_nv{(uint8_t *)name, (uint8_t *)value.data(), N - 1,
-                    value.size(), NGHTTP3_NV_FLAG_NONE};
+inline nghttp3_nv make_nv_cc(const std::string_view &name,
+                             const std::string_view &value) {
+  return make_nv(name, value, NGHTTP3_NV_FLAG_NONE);
 }
 
-template <typename S1, typename S2>
-constexpr nghttp3_nv make_nv(const S1 &name, const S2 &value) {
-  return nghttp3_nv{(uint8_t *)name.data(), (uint8_t *)value.data(),
-                    name.size(), value.size(), NGHTTP3_NV_FLAG_NONE};
+inline nghttp3_nv make_nv_nc(const std::string_view &name,
+                             const std::string_view &value) {
+  return make_nv(name, value, NGHTTP3_NV_FLAG_NO_COPY_NAME);
+}
+
+inline nghttp3_nv make_nv_nn(const std::string_view &name,
+                             const std::string_view &value) {
+  return make_nv(name, value,
+                 NGHTTP3_NV_FLAG_NO_COPY_NAME | NGHTTP3_NV_FLAG_NO_COPY_VALUE);
 }
 
 std::string format_hex(uint8_t c);
 
 std::string format_hex(const uint8_t *s, size_t len);
 
-std::string format_hex(const std::string &s);
+std::string format_hex(const std::string_view &s);
 
 template <size_t N> std::string format_hex(const uint8_t (&s)[N]) {
   return format_hex(s, N);
 }
 
-std::string decode_hex(const std::string &s);
+std::string decode_hex(const std::string_view &s);
 
 // format_durationf formats |ns| in human readable manner.  |ns| must
 // be nanoseconds resolution.  This function uses the largest unit so
@@ -138,11 +147,6 @@ template <typename S, typename T> bool istarts_with(const S &a, const T &b) {
   return istarts_with(a.begin(), a.end(), b.begin(), b.end());
 }
 
-template <typename T, typename CharT, size_t N>
-bool istarts_with_l(const T &a, const CharT (&b)[N]) {
-  return istarts_with(a.begin(), a.end(), b, b + N - 1);
-}
-
 // make_cid_key returns the key for |cid|.
 std::string make_cid_key(const ngtcp2_cid *cid);
 std::string make_cid_key(const uint8_t *cid, size_t cidlen);
@@ -279,7 +283,7 @@ int generate_secret(uint8_t *secret, size_t secretlen);
 // normalize_path removes ".." by consuming a previous path component.
 // It also removes ".".  It assumes that |path| starts with "/".  If
 // it cannot consume a previous path component, it just removes "..".
-std::string normalize_path(const std::string &path);
+std::string normalize_path(const std::string_view &path);
 
 constexpr bool is_digit(const char c) { return '0' <= c && c <= '9'; }
 
@@ -338,6 +342,16 @@ const char *crypto_default_ciphers();
 
 const char *crypto_default_groups();
 
+// split_str parses delimited strings in |s| and returns substrings
+// delimited by |delim|.  The any white spaces around substring are
+// treated as a part of substring.
+std::vector<std::string_view> split_str(const std::string_view &s,
+                                        char delim = ',');
+
+// parse_version parses |s| to get 4 byte QUIC version.  |s| must be a
+// hex string and must start with "0x" (.e.g, 0x00000001).
+std::optional<uint32_t> parse_version(const std::string_view &s);
+
 } // namespace util
 
 std::ostream &operator<<(std::ostream &os, const ngtcp2_cid &cid);
