@@ -47,6 +47,7 @@
 #include "ngtcp2_vec_test.h"
 #include "ngtcp2_strm_test.h"
 #include "ngtcp2_pv_test.h"
+#include "ngtcp2_pmtud_test.h"
 #include "ngtcp2_str_test.h"
 
 static int init_suite1(void) { return 0; }
@@ -138,8 +139,14 @@ int main(void) {
       !CU_add_test(pSuite, "pkt_stream_max_datalen",
                    test_ngtcp2_pkt_stream_max_datalen) ||
       !CU_add_test(pSuite, "get_varint", test_ngtcp2_get_varint) ||
-      !CU_add_test(pSuite, "get_varint_len", test_ngtcp2_get_varint_len) ||
-      !CU_add_test(pSuite, "put_varint_len", test_ngtcp2_put_varint_len) ||
+      !CU_add_test(pSuite, "get_uvarintlen", test_ngtcp2_get_uvarintlen) ||
+      !CU_add_test(pSuite, "put_uvarintlen", test_ngtcp2_put_uvarintlen) ||
+      !CU_add_test(pSuite, "get_uint64", test_ngtcp2_get_uint64) ||
+      !CU_add_test(pSuite, "get_uint48", test_ngtcp2_get_uint48) ||
+      !CU_add_test(pSuite, "get_uint32", test_ngtcp2_get_uint32) ||
+      !CU_add_test(pSuite, "get_uint24", test_ngtcp2_get_uint24) ||
+      !CU_add_test(pSuite, "get_uint16", test_ngtcp2_get_uint16) ||
+      !CU_add_test(pSuite, "get_uint16be", test_ngtcp2_get_uint16be) ||
       !CU_add_test(pSuite, "nth_server_bidi_id",
                    test_ngtcp2_nth_server_bidi_id) ||
       !CU_add_test(pSuite, "nth_server_uni_id",
@@ -169,6 +176,8 @@ int main(void) {
       !CU_add_test(pSuite, "acktr_recv_ack", test_ngtcp2_acktr_recv_ack) ||
       !CU_add_test(pSuite, "encode_transport_params",
                    test_ngtcp2_encode_transport_params) ||
+      !CU_add_test(pSuite, "decode_transport_params_new",
+                   test_ngtcp2_decode_transport_params_new) ||
       !CU_add_test(pSuite, "rtb_add", test_ngtcp2_rtb_add) ||
       !CU_add_test(pSuite, "rtb_recv_ack", test_ngtcp2_rtb_recv_ack) ||
       !CU_add_test(pSuite, "rtb_lost_pkt_ts", test_ngtcp2_rtb_lost_pkt_ts) ||
@@ -294,7 +303,17 @@ int main(void) {
                    test_ngtcp2_conn_handshake_timeout) ||
       !CU_add_test(pSuite, "conn_get_connection_close_error",
                    test_ngtcp2_conn_get_connection_close_error) ||
+      !CU_add_test(pSuite, "conn_version_negotiation",
+                   test_ngtcp2_conn_version_negotiation) ||
+      !CU_add_test(pSuite, "conn_server_negotiate_version",
+                   test_ngtcp2_conn_server_negotiate_version) ||
+      !CU_add_test(pSuite, "conn_pmtud_loss", test_ngtcp2_conn_pmtud_loss) ||
+      !CU_add_test(pSuite, "conn_amplification",
+                   test_ngtcp2_conn_amplification) ||
+      !CU_add_test(pSuite, "conn_new_failmalloc",
+                   test_ngtcp2_conn_new_failmalloc) ||
       !CU_add_test(pSuite, "accept", test_ngtcp2_accept) ||
+      !CU_add_test(pSuite, "select_version", test_ngtcp2_select_version) ||
       !CU_add_test(pSuite, "pkt_write_connection_close",
                    test_ngtcp2_pkt_write_connection_close) ||
       !CU_add_test(pSuite, "map", test_ngtcp2_map) ||
@@ -316,8 +335,10 @@ int main(void) {
                    test_ngtcp2_strm_streamfrq_unacked_pop) ||
       !CU_add_test(pSuite, "pv_add_entry", test_ngtcp2_pv_add_entry) ||
       !CU_add_test(pSuite, "pv_validate", test_ngtcp2_pv_validate) ||
+      !CU_add_test(pSuite, "pmtud_probe", test_ngtcp2_pmtud_probe) ||
       !CU_add_test(pSuite, "encode_ipv4", test_ngtcp2_encode_ipv4) ||
-      !CU_add_test(pSuite, "encode_ipv6", test_ngtcp2_encode_ipv6)) {
+      !CU_add_test(pSuite, "encode_ipv6", test_ngtcp2_encode_ipv6) ||
+      !CU_add_test(pSuite, "get_bytes", test_ngtcp2_get_bytes)) {
     CU_cleanup_registry();
     return (int)CU_get_error();
   }
