@@ -57,33 +57,49 @@ static int null_retry_encrypt(uint8_t *dest, const ngtcp2_crypto_aead *aead,
 }
 
 void test_ngtcp2_pkt_decode_version_cid(void) {
-  uint8_t buf[1024];
-  uint32_t version;
-  const uint8_t *dcid, *scid;
-  size_t dcidlen, scidlen;
+  uint8_t buf[NGTCP2_MAX_UDP_PAYLOAD_SIZE];
+  ngtcp2_version_cid vc;
   int rv;
   uint8_t *p;
 
   /* Supported QUIC version */
   p = buf;
   *p++ = NGTCP2_HEADER_FORM_BIT;
-  p = ngtcp2_put_uint32be(p, NGTCP2_PROTO_VER_MAX);
+  p = ngtcp2_put_uint32be(p, NGTCP2_PROTO_VER_V1);
   *p++ = NGTCP2_MAX_CIDLEN;
   p = ngtcp2_setmem(p, 0xf1, NGTCP2_MAX_CIDLEN);
   *p++ = NGTCP2_MAX_CIDLEN - 1;
   p = ngtcp2_setmem(p, 0xf2, NGTCP2_MAX_CIDLEN - 1);
 
-  rv = ngtcp2_pkt_decode_version_cid(&version, &dcid, &dcidlen, &scid, &scidlen,
-                                     buf, (size_t)(p - buf), 0);
+  rv = ngtcp2_pkt_decode_version_cid(&vc, buf, (size_t)(p - buf), 0);
 
   CU_ASSERT(0 == rv);
-  CU_ASSERT(NGTCP2_PROTO_VER_MAX == version);
-  CU_ASSERT(NGTCP2_MAX_CIDLEN == dcidlen);
-  CU_ASSERT(&buf[6] == dcid);
-  CU_ASSERT(NGTCP2_MAX_CIDLEN - 1 == scidlen);
-  CU_ASSERT(&buf[6 + NGTCP2_MAX_CIDLEN + 1] == scid);
+  CU_ASSERT(NGTCP2_PROTO_VER_V1 == vc.version);
+  CU_ASSERT(NGTCP2_MAX_CIDLEN == vc.dcidlen);
+  CU_ASSERT(&buf[6] == vc.dcid);
+  CU_ASSERT(NGTCP2_MAX_CIDLEN - 1 == vc.scidlen);
+  CU_ASSERT(&buf[6 + NGTCP2_MAX_CIDLEN + 1] == vc.scid);
 
   /* Unsupported QUIC version */
+  memset(buf, 0, sizeof(buf));
+  p = buf;
+  *p++ = NGTCP2_HEADER_FORM_BIT;
+  p = ngtcp2_put_uint32be(p, 0xffffff00);
+  *p++ = NGTCP2_MAX_CIDLEN;
+  p = ngtcp2_setmem(p, 0xf1, NGTCP2_MAX_CIDLEN);
+  *p++ = NGTCP2_MAX_CIDLEN - 1;
+  ngtcp2_setmem(p, 0xf2, NGTCP2_MAX_CIDLEN - 1);
+
+  rv = ngtcp2_pkt_decode_version_cid(&vc, buf, sizeof(buf), 0);
+
+  CU_ASSERT(NGTCP2_ERR_VERSION_NEGOTIATION == rv);
+  CU_ASSERT(0xffffff00 == vc.version);
+  CU_ASSERT(NGTCP2_MAX_CIDLEN == vc.dcidlen);
+  CU_ASSERT(&buf[6] == vc.dcid);
+  CU_ASSERT(NGTCP2_MAX_CIDLEN - 1 == vc.scidlen);
+  CU_ASSERT(&buf[6 + NGTCP2_MAX_CIDLEN + 1] == vc.scid);
+
+  /* Unsupported QUIC version with UDP payload size < 1200 */
   p = buf;
   *p++ = NGTCP2_HEADER_FORM_BIT;
   p = ngtcp2_put_uint32be(p, 0xffffff00);
@@ -92,48 +108,41 @@ void test_ngtcp2_pkt_decode_version_cid(void) {
   *p++ = NGTCP2_MAX_CIDLEN - 1;
   p = ngtcp2_setmem(p, 0xf2, NGTCP2_MAX_CIDLEN - 1);
 
-  rv = ngtcp2_pkt_decode_version_cid(&version, &dcid, &dcidlen, &scid, &scidlen,
-                                     buf, (size_t)(p - buf), 0);
+  rv = ngtcp2_pkt_decode_version_cid(&vc, buf, (size_t)(p - buf), 0);
 
-  CU_ASSERT(NGTCP2_ERR_VERSION_NEGOTIATION == rv);
-  CU_ASSERT(0xffffff00 == version);
-  CU_ASSERT(NGTCP2_MAX_CIDLEN == dcidlen);
-  CU_ASSERT(&buf[6] == dcid);
-  CU_ASSERT(NGTCP2_MAX_CIDLEN - 1 == scidlen);
-  CU_ASSERT(&buf[6 + NGTCP2_MAX_CIDLEN + 1] == scid);
+  CU_ASSERT(NGTCP2_ERR_INVALID_ARGUMENT == rv);
 
   /* Supported QUIC version with long CID */
   p = buf;
   *p++ = NGTCP2_HEADER_FORM_BIT;
-  p = ngtcp2_put_uint32be(p, NGTCP2_PROTO_VER_MAX);
+  p = ngtcp2_put_uint32be(p, NGTCP2_PROTO_VER_V1);
   *p++ = NGTCP2_MAX_CIDLEN + 1;
   p = ngtcp2_setmem(p, 0xf1, NGTCP2_MAX_CIDLEN + 1);
   *p++ = NGTCP2_MAX_CIDLEN;
   p = ngtcp2_setmem(p, 0xf2, NGTCP2_MAX_CIDLEN);
 
-  rv = ngtcp2_pkt_decode_version_cid(&version, &dcid, &dcidlen, &scid, &scidlen,
-                                     buf, (size_t)(p - buf), 0);
+  rv = ngtcp2_pkt_decode_version_cid(&vc, buf, (size_t)(p - buf), 0);
 
   CU_ASSERT(NGTCP2_ERR_INVALID_ARGUMENT == rv);
 
   /* Unsupported QUIC version with long CID */
+  memset(buf, 0, sizeof(buf));
   p = buf;
   *p++ = NGTCP2_HEADER_FORM_BIT;
   p = ngtcp2_put_uint32be(p, 0xffffff00);
   *p++ = NGTCP2_MAX_CIDLEN + 1;
   p = ngtcp2_setmem(p, 0xf1, NGTCP2_MAX_CIDLEN + 1);
   *p++ = NGTCP2_MAX_CIDLEN;
-  p = ngtcp2_setmem(p, 0xf2, NGTCP2_MAX_CIDLEN);
+  ngtcp2_setmem(p, 0xf2, NGTCP2_MAX_CIDLEN);
 
-  rv = ngtcp2_pkt_decode_version_cid(&version, &dcid, &dcidlen, &scid, &scidlen,
-                                     buf, (size_t)(p - buf), 0);
+  rv = ngtcp2_pkt_decode_version_cid(&vc, buf, sizeof(buf), 0);
 
   CU_ASSERT(NGTCP2_ERR_VERSION_NEGOTIATION == rv);
-  CU_ASSERT(0xffffff00 == version);
-  CU_ASSERT(NGTCP2_MAX_CIDLEN + 1 == dcidlen);
-  CU_ASSERT(&buf[6] == dcid);
-  CU_ASSERT(NGTCP2_MAX_CIDLEN == scidlen);
-  CU_ASSERT(&buf[6 + NGTCP2_MAX_CIDLEN + 1 + 1] == scid);
+  CU_ASSERT(0xffffff00 == vc.version);
+  CU_ASSERT(NGTCP2_MAX_CIDLEN + 1 == vc.dcidlen);
+  CU_ASSERT(&buf[6] == vc.dcid);
+  CU_ASSERT(NGTCP2_MAX_CIDLEN == vc.scidlen);
+  CU_ASSERT(&buf[6 + NGTCP2_MAX_CIDLEN + 1 + 1] == vc.scid);
 
   /* VN */
   p = buf;
@@ -144,15 +153,14 @@ void test_ngtcp2_pkt_decode_version_cid(void) {
   *p++ = NGTCP2_MAX_CIDLEN - 1;
   p = ngtcp2_setmem(p, 0xf2, NGTCP2_MAX_CIDLEN - 1);
 
-  rv = ngtcp2_pkt_decode_version_cid(&version, &dcid, &dcidlen, &scid, &scidlen,
-                                     buf, (size_t)(p - buf), 0);
+  rv = ngtcp2_pkt_decode_version_cid(&vc, buf, (size_t)(p - buf), 0);
 
   CU_ASSERT(0 == rv);
-  CU_ASSERT(0 == version);
-  CU_ASSERT(NGTCP2_MAX_CIDLEN == dcidlen);
-  CU_ASSERT(&buf[6] == dcid);
-  CU_ASSERT(NGTCP2_MAX_CIDLEN - 1 == scidlen);
-  CU_ASSERT(&buf[6 + NGTCP2_MAX_CIDLEN + 1] == scid);
+  CU_ASSERT(0 == vc.version);
+  CU_ASSERT(NGTCP2_MAX_CIDLEN == vc.dcidlen);
+  CU_ASSERT(&buf[6] == vc.dcid);
+  CU_ASSERT(NGTCP2_MAX_CIDLEN - 1 == vc.scidlen);
+  CU_ASSERT(&buf[6 + NGTCP2_MAX_CIDLEN + 1] == vc.scid);
 
   /* VN with long CID */
   p = buf;
@@ -163,23 +171,26 @@ void test_ngtcp2_pkt_decode_version_cid(void) {
   *p++ = NGTCP2_MAX_CIDLEN;
   p = ngtcp2_setmem(p, 0xf2, NGTCP2_MAX_CIDLEN);
 
-  rv = ngtcp2_pkt_decode_version_cid(&version, &dcid, &dcidlen, &scid, &scidlen,
-                                     buf, (size_t)(p - buf), 0);
+  rv = ngtcp2_pkt_decode_version_cid(&vc, buf, (size_t)(p - buf), 0);
 
-  CU_ASSERT(NGTCP2_ERR_INVALID_ARGUMENT == rv);
+  CU_ASSERT(0 == rv);
+  CU_ASSERT(0 == vc.version);
+  CU_ASSERT(NGTCP2_MAX_CIDLEN + 1 == vc.dcidlen);
+  CU_ASSERT(&buf[6] == vc.dcid);
+  CU_ASSERT(NGTCP2_MAX_CIDLEN == vc.scidlen);
+  CU_ASSERT(&buf[6 + NGTCP2_MAX_CIDLEN + 1 + 1] == vc.scid);
 
   /* Malformed Long packet */
   p = buf;
   *p++ = NGTCP2_HEADER_FORM_BIT;
-  p = ngtcp2_put_uint32be(p, NGTCP2_PROTO_VER_MAX);
+  p = ngtcp2_put_uint32be(p, NGTCP2_PROTO_VER_V1);
   *p++ = NGTCP2_MAX_CIDLEN;
   p = ngtcp2_setmem(p, 0xf1, NGTCP2_MAX_CIDLEN);
   *p++ = NGTCP2_MAX_CIDLEN - 1;
   p = ngtcp2_setmem(p, 0xf2, NGTCP2_MAX_CIDLEN - 1);
   --p;
 
-  rv = ngtcp2_pkt_decode_version_cid(&version, &dcid, &dcidlen, &scid, &scidlen,
-                                     buf, (size_t)(p - buf), 0);
+  rv = ngtcp2_pkt_decode_version_cid(&vc, buf, (size_t)(p - buf), 0);
 
   CU_ASSERT(NGTCP2_ERR_INVALID_ARGUMENT == rv);
 
@@ -188,15 +199,15 @@ void test_ngtcp2_pkt_decode_version_cid(void) {
   *p++ = 0;
   p = ngtcp2_setmem(p, 0xf1, NGTCP2_MAX_CIDLEN);
 
-  rv = ngtcp2_pkt_decode_version_cid(&version, &dcid, &dcidlen, &scid, &scidlen,
-                                     buf, (size_t)(p - buf), NGTCP2_MAX_CIDLEN);
+  rv = ngtcp2_pkt_decode_version_cid(&vc, buf, (size_t)(p - buf),
+                                     NGTCP2_MAX_CIDLEN);
 
   CU_ASSERT(0 == rv);
-  CU_ASSERT(0 == version);
-  CU_ASSERT(&buf[1] == dcid);
-  CU_ASSERT(NGTCP2_MAX_CIDLEN == dcidlen);
-  CU_ASSERT(NULL == scid);
-  CU_ASSERT(0 == scidlen);
+  CU_ASSERT(0 == vc.version);
+  CU_ASSERT(&buf[1] == vc.dcid);
+  CU_ASSERT(NGTCP2_MAX_CIDLEN == vc.dcidlen);
+  CU_ASSERT(NULL == vc.scid);
+  CU_ASSERT(0 == vc.scidlen);
 
   /* Malformed Short packet */
   p = buf;
@@ -204,8 +215,8 @@ void test_ngtcp2_pkt_decode_version_cid(void) {
   p = ngtcp2_setmem(p, 0xf1, NGTCP2_MAX_CIDLEN);
   --p;
 
-  rv = ngtcp2_pkt_decode_version_cid(&version, &dcid, &dcidlen, &scid, &scidlen,
-                                     buf, (size_t)(p - buf), NGTCP2_MAX_CIDLEN);
+  rv = ngtcp2_pkt_decode_version_cid(&vc, buf, (size_t)(p - buf),
+                                     NGTCP2_MAX_CIDLEN);
 
   CU_ASSERT(NGTCP2_ERR_INVALID_ARGUMENT == rv);
 }
@@ -222,7 +233,7 @@ void test_ngtcp2_pkt_decode_hd_long(void) {
 
   /* Handshake */
   ngtcp2_pkt_hd_init(&hd, NGTCP2_PKT_FLAG_LONG_FORM, NGTCP2_PKT_HANDSHAKE,
-                     &dcid, &scid, 0xe1e2e3e4u, 4, 0x000000ff, 16383);
+                     &dcid, &scid, 0xe1e2e3e4u, 4, NGTCP2_PROTO_VER_V1, 16383);
 
   rv = ngtcp2_pkt_encode_hd_long(buf, sizeof(buf), &hd);
 
@@ -245,7 +256,8 @@ void test_ngtcp2_pkt_decode_hd_long(void) {
   /* Handshake without Fixed Bit set */
   ngtcp2_pkt_hd_init(
       &hd, NGTCP2_PKT_FLAG_LONG_FORM | NGTCP2_PKT_FLAG_FIXED_BIT_CLEAR,
-      NGTCP2_PKT_HANDSHAKE, &dcid, &scid, 0xe1e2e3e4u, 4, 0x000000ff, 16383);
+      NGTCP2_PKT_HANDSHAKE, &dcid, &scid, 0xe1e2e3e4u, 4, NGTCP2_PROTO_VER_V1,
+      16383);
 
   rv = ngtcp2_pkt_encode_hd_long(buf, sizeof(buf), &hd);
 
@@ -268,9 +280,11 @@ void test_ngtcp2_pkt_decode_hd_long(void) {
   /* VN */
   /* Set random packet type */
   ngtcp2_pkt_hd_init(&hd, NGTCP2_PKT_FLAG_LONG_FORM, NGTCP2_PKT_HANDSHAKE,
-                     &dcid, &scid, 0, 4, 0, 0);
+                     &dcid, &scid, 0, 4, NGTCP2_PROTO_VER_V1, 0);
 
   rv = ngtcp2_pkt_encode_hd_long(buf, sizeof(buf), &hd);
+  /* Set version field to 0 */
+  memset(&buf[1], 0, 4);
 
   len = 1 + 4 + 1 + dcid.datalen + 1 + scid.datalen;
 
@@ -284,7 +298,7 @@ void test_ngtcp2_pkt_decode_hd_long(void) {
   CU_ASSERT(ngtcp2_cid_eq(&hd.dcid, &nhd.dcid));
   CU_ASSERT(ngtcp2_cid_eq(&hd.scid, &nhd.scid));
   CU_ASSERT(hd.pkt_num == nhd.pkt_num);
-  CU_ASSERT(hd.version == nhd.version);
+  CU_ASSERT(0 == nhd.version);
   CU_ASSERT(hd.len == nhd.len);
 }
 
@@ -299,7 +313,7 @@ void test_ngtcp2_pkt_decode_hd_short(void) {
   ngtcp2_cid_zero(&zcid);
 
   /* 4 bytes packet number */
-  ngtcp2_pkt_hd_init(&hd, NGTCP2_PKT_FLAG_NONE, NGTCP2_PKT_SHORT, &dcid, NULL,
+  ngtcp2_pkt_hd_init(&hd, NGTCP2_PKT_FLAG_NONE, NGTCP2_PKT_1RTT, &dcid, NULL,
                      0xe1e2e3e4u, 4, 0xd1d2d3d4u, 0);
 
   expectedlen = 1 + dcid.datalen + 4;
@@ -313,7 +327,7 @@ void test_ngtcp2_pkt_decode_hd_short(void) {
 
   CU_ASSERT((ngtcp2_ssize)expectedlen == rv);
   CU_ASSERT(hd.flags == nhd.flags);
-  CU_ASSERT(NGTCP2_PKT_SHORT == nhd.type);
+  CU_ASSERT(NGTCP2_PKT_1RTT == nhd.type);
   CU_ASSERT(ngtcp2_cid_eq(&dcid, &nhd.dcid));
   CU_ASSERT(ngtcp2_cid_empty(&nhd.scid));
   CU_ASSERT(0xe1e2e3e4u == nhd.pkt_num);
@@ -324,7 +338,7 @@ void test_ngtcp2_pkt_decode_hd_short(void) {
   /* 4 bytes packet number without Fixed Bit set */
   ngtcp2_pkt_hd_init(
       &hd, NGTCP2_PKT_FLAG_NONE | NGTCP2_PKT_FLAG_FIXED_BIT_CLEAR,
-      NGTCP2_PKT_SHORT, &dcid, NULL, 0xe1e2e3e4u, 4, 0xd1d2d3d4u, 0);
+      NGTCP2_PKT_1RTT, &dcid, NULL, 0xe1e2e3e4u, 4, 0xd1d2d3d4u, 0);
 
   expectedlen = 1 + dcid.datalen + 4;
 
@@ -337,7 +351,7 @@ void test_ngtcp2_pkt_decode_hd_short(void) {
 
   CU_ASSERT((ngtcp2_ssize)expectedlen == rv);
   CU_ASSERT(hd.flags == nhd.flags);
-  CU_ASSERT(NGTCP2_PKT_SHORT == nhd.type);
+  CU_ASSERT(NGTCP2_PKT_1RTT == nhd.type);
   CU_ASSERT(ngtcp2_cid_eq(&dcid, &nhd.dcid));
   CU_ASSERT(ngtcp2_cid_empty(&nhd.scid));
   CU_ASSERT(0xe1e2e3e4u == nhd.pkt_num);
@@ -346,7 +360,7 @@ void test_ngtcp2_pkt_decode_hd_short(void) {
   CU_ASSERT(0 == nhd.len);
 
   /* 2 bytes packet number */
-  ngtcp2_pkt_hd_init(&hd, NGTCP2_PKT_FLAG_NONE, NGTCP2_PKT_SHORT, &dcid, NULL,
+  ngtcp2_pkt_hd_init(&hd, NGTCP2_PKT_FLAG_NONE, NGTCP2_PKT_1RTT, &dcid, NULL,
                      0xe1e2e3e4u, 2, 0xd1d2d3d4u, 0);
 
   expectedlen = 1 + dcid.datalen + 2;
@@ -359,7 +373,7 @@ void test_ngtcp2_pkt_decode_hd_short(void) {
 
   CU_ASSERT((ngtcp2_ssize)expectedlen == rv);
   CU_ASSERT(hd.flags == nhd.flags);
-  CU_ASSERT(NGTCP2_PKT_SHORT == nhd.type);
+  CU_ASSERT(NGTCP2_PKT_1RTT == nhd.type);
   CU_ASSERT(ngtcp2_cid_eq(&dcid, &nhd.dcid));
   CU_ASSERT(ngtcp2_cid_empty(&nhd.scid));
   CU_ASSERT(0xe3e4u == nhd.pkt_num);
@@ -368,7 +382,7 @@ void test_ngtcp2_pkt_decode_hd_short(void) {
   CU_ASSERT(0 == nhd.len);
 
   /* 1 byte packet number */
-  ngtcp2_pkt_hd_init(&hd, NGTCP2_PKT_FLAG_NONE, NGTCP2_PKT_SHORT, &dcid, NULL,
+  ngtcp2_pkt_hd_init(&hd, NGTCP2_PKT_FLAG_NONE, NGTCP2_PKT_1RTT, &dcid, NULL,
                      0xe1e2e3e4u, 1, 0xd1d2d3d4u, 0);
 
   expectedlen = 1 + dcid.datalen + 1;
@@ -381,7 +395,7 @@ void test_ngtcp2_pkt_decode_hd_short(void) {
 
   CU_ASSERT((ngtcp2_ssize)expectedlen == rv);
   CU_ASSERT(hd.flags == nhd.flags);
-  CU_ASSERT(NGTCP2_PKT_SHORT == nhd.type);
+  CU_ASSERT(NGTCP2_PKT_1RTT == nhd.type);
   CU_ASSERT(ngtcp2_cid_eq(&dcid, &nhd.dcid));
   CU_ASSERT(ngtcp2_cid_empty(&nhd.scid));
   CU_ASSERT(0xe4 == nhd.pkt_num);
@@ -390,7 +404,7 @@ void test_ngtcp2_pkt_decode_hd_short(void) {
   CU_ASSERT(0 == nhd.len);
 
   /* With Key Phase */
-  ngtcp2_pkt_hd_init(&hd, NGTCP2_PKT_FLAG_KEY_PHASE, NGTCP2_PKT_SHORT, &dcid,
+  ngtcp2_pkt_hd_init(&hd, NGTCP2_PKT_FLAG_KEY_PHASE, NGTCP2_PKT_1RTT, &dcid,
                      NULL, 0xe1e2e3e4u, 4, 0xd1d2d3d4u, 0);
 
   expectedlen = 1 + dcid.datalen + 4;
@@ -405,7 +419,7 @@ void test_ngtcp2_pkt_decode_hd_short(void) {
   /* key phase bit is protected by header protection and
      ngtcp2_pkt_decode_hd_short does not decode it. */
   CU_ASSERT(NGTCP2_PKT_FLAG_NONE == nhd.flags);
-  CU_ASSERT(NGTCP2_PKT_SHORT == nhd.type);
+  CU_ASSERT(NGTCP2_PKT_1RTT == nhd.type);
   CU_ASSERT(ngtcp2_cid_eq(&dcid, &nhd.dcid));
   CU_ASSERT(ngtcp2_cid_empty(&nhd.scid));
   CU_ASSERT(0xe1e2e3e4u == nhd.pkt_num);
@@ -414,7 +428,7 @@ void test_ngtcp2_pkt_decode_hd_short(void) {
   CU_ASSERT(0 == nhd.len);
 
   /* With empty DCID */
-  ngtcp2_pkt_hd_init(&hd, NGTCP2_PKT_FLAG_NONE, NGTCP2_PKT_SHORT, NULL, NULL,
+  ngtcp2_pkt_hd_init(&hd, NGTCP2_PKT_FLAG_NONE, NGTCP2_PKT_1RTT, NULL, NULL,
                      0xe1e2e3e4u, 4, 0xd1d2d3d4u, 0);
 
   expectedlen = 1 + 4;
@@ -427,7 +441,7 @@ void test_ngtcp2_pkt_decode_hd_short(void) {
 
   CU_ASSERT((ngtcp2_ssize)expectedlen == rv);
   CU_ASSERT(hd.flags == nhd.flags);
-  CU_ASSERT(NGTCP2_PKT_SHORT == nhd.type);
+  CU_ASSERT(NGTCP2_PKT_1RTT == nhd.type);
   CU_ASSERT(ngtcp2_cid_empty(&nhd.dcid));
   CU_ASSERT(ngtcp2_cid_empty(&nhd.scid));
   CU_ASSERT(0xe1e2e3e4u == nhd.pkt_num);
@@ -533,16 +547,16 @@ void test_ngtcp2_pkt_decode_ack_frame(void) {
 
   CU_ASSERT((ngtcp2_ssize)expectedlen == rv);
   CU_ASSERT(0x31f2f3f4f5f6f7f8llu == fr.ack.largest_ack);
-  CU_ASSERT(1 == fr.ack.num_blks);
-  CU_ASSERT(0x31e2e3e4e5e6e7e8llu == fr.ack.first_ack_blklen);
-  CU_ASSERT(99 == fr.ack.blks[0].gap);
-  CU_ASSERT(0x31d2d3d4d5d6d7d8llu == fr.ack.blks[0].blklen);
+  CU_ASSERT(1 == fr.ack.rangecnt);
+  CU_ASSERT(0x31e2e3e4e5e6e7e8llu == fr.ack.first_ack_range);
+  CU_ASSERT(99 == fr.ack.ranges[0].gap);
+  CU_ASSERT(0x31d2d3d4d5d6d7d8llu == fr.ack.ranges[0].len);
 }
 
 void test_ngtcp2_pkt_decode_padding_frame(void) {
   uint8_t buf[256];
   ngtcp2_frame fr;
-  size_t rv;
+  ngtcp2_ssize rv;
   size_t paddinglen = 31;
 
   memset(buf, 0, paddinglen);
@@ -550,7 +564,7 @@ void test_ngtcp2_pkt_decode_padding_frame(void) {
 
   rv = ngtcp2_pkt_decode_padding_frame(&fr.padding, buf, paddinglen + 1);
 
-  CU_ASSERT(paddinglen == rv);
+  CU_ASSERT((ngtcp2_ssize)paddinglen == rv);
   CU_ASSERT((size_t)31 == fr.padding.len);
 }
 
@@ -684,14 +698,14 @@ void test_ngtcp2_pkt_encode_ack_frame(void) {
   ngtcp2_ssize rv;
   size_t framelen;
   size_t i;
-  ngtcp2_ack_blk *blks;
+  ngtcp2_ack_range *ranges;
 
   /* 0 Num Blocks */
   fr->type = NGTCP2_FRAME_ACK;
   fr->ack.largest_ack = 0xf1f2f3f4llu;
-  fr->ack.first_ack_blklen = 0;
+  fr->ack.first_ack_range = 0;
   fr->ack.ack_delay = 0;
-  fr->ack.num_blks = 0;
+  fr->ack.rangecnt = 0;
 
   framelen = 1 + 8 + 1 + 1 + 1;
 
@@ -705,21 +719,21 @@ void test_ngtcp2_pkt_encode_ack_frame(void) {
   CU_ASSERT(fr->type == nfr->type);
   CU_ASSERT(fr->ack.largest_ack == nfr->ack.largest_ack);
   CU_ASSERT(fr->ack.ack_delay == nfr->ack.ack_delay);
-  CU_ASSERT(fr->ack.num_blks == nfr->ack.num_blks);
+  CU_ASSERT(fr->ack.rangecnt == nfr->ack.rangecnt);
 
   memset(&nmfr, 0, sizeof(nmfr));
 
   /* 2 Num Blocks */
   fr->type = NGTCP2_FRAME_ACK;
   fr->ack.largest_ack = 0xf1f2f3f4llu;
-  fr->ack.first_ack_blklen = 0xe1e2e3e4llu;
+  fr->ack.first_ack_range = 0xe1e2e3e4llu;
   fr->ack.ack_delay = 0xf1f2;
-  fr->ack.num_blks = 2;
-  blks = fr->ack.blks;
-  blks[0].gap = 255;
-  blks[0].blklen = 0xd1d2d3d4llu;
-  blks[1].gap = 1;
-  blks[1].blklen = 0xd1d2d3d4llu;
+  fr->ack.rangecnt = 2;
+  ranges = fr->ack.ranges;
+  ranges[0].gap = 255;
+  ranges[0].len = 0xd1d2d3d4llu;
+  ranges[1].gap = 1;
+  ranges[1].len = 0xd1d2d3d4llu;
 
   framelen = 1 + 8 + 4 + 1 + 8 + (2 + 8) + (1 + 8);
 
@@ -733,11 +747,11 @@ void test_ngtcp2_pkt_encode_ack_frame(void) {
   CU_ASSERT(fr->type == nfr->type);
   CU_ASSERT(fr->ack.largest_ack == nfr->ack.largest_ack);
   CU_ASSERT(fr->ack.ack_delay == nfr->ack.ack_delay);
-  CU_ASSERT(fr->ack.num_blks == nfr->ack.num_blks);
+  CU_ASSERT(fr->ack.rangecnt == nfr->ack.rangecnt);
 
-  for (i = 0; i < fr->ack.num_blks; ++i) {
-    CU_ASSERT(fr->ack.blks[i].gap == nfr->ack.blks[i].gap);
-    CU_ASSERT(fr->ack.blks[i].blklen == nfr->ack.blks[i].blklen);
+  for (i = 0; i < fr->ack.rangecnt; ++i) {
+    CU_ASSERT(fr->ack.ranges[i].gap == nfr->ack.ranges[i].gap);
+    CU_ASSERT(fr->ack.ranges[i].len == nfr->ack.ranges[i].len);
   }
 
   memset(&nmfr, 0, sizeof(nmfr));
@@ -750,14 +764,14 @@ void test_ngtcp2_pkt_encode_ack_ecn_frame(void) {
   ngtcp2_ssize rv;
   size_t framelen;
   size_t i;
-  ngtcp2_ack_blk *blks;
+  ngtcp2_ack_range *ranges;
 
   /* 0 Num Blocks */
   fr->type = NGTCP2_FRAME_ACK_ECN;
   fr->ack.largest_ack = 0xf1f2f3f4llu;
-  fr->ack.first_ack_blklen = 0;
+  fr->ack.first_ack_range = 0;
   fr->ack.ack_delay = 0;
-  fr->ack.num_blks = 0;
+  fr->ack.rangecnt = 0;
   fr->ack.ecn.ect0 = 64;
   fr->ack.ecn.ect1 = 16384;
   fr->ack.ecn.ce = 1073741824;
@@ -774,7 +788,7 @@ void test_ngtcp2_pkt_encode_ack_ecn_frame(void) {
   CU_ASSERT(fr->type == nfr->type);
   CU_ASSERT(fr->ack.largest_ack == nfr->ack.largest_ack);
   CU_ASSERT(fr->ack.ack_delay == nfr->ack.ack_delay);
-  CU_ASSERT(fr->ack.num_blks == nfr->ack.num_blks);
+  CU_ASSERT(fr->ack.rangecnt == nfr->ack.rangecnt);
   CU_ASSERT(fr->ack.ecn.ect0 == nfr->ack.ecn.ect0);
   CU_ASSERT(fr->ack.ecn.ect1 == nfr->ack.ecn.ect1);
   CU_ASSERT(fr->ack.ecn.ce == nfr->ack.ecn.ce);
@@ -784,14 +798,14 @@ void test_ngtcp2_pkt_encode_ack_ecn_frame(void) {
   /* 2 Num Blocks */
   fr->type = NGTCP2_FRAME_ACK_ECN;
   fr->ack.largest_ack = 0xf1f2f3f4llu;
-  fr->ack.first_ack_blklen = 0xe1e2e3e4llu;
+  fr->ack.first_ack_range = 0xe1e2e3e4llu;
   fr->ack.ack_delay = 0xf1f2;
-  fr->ack.num_blks = 2;
-  blks = fr->ack.blks;
-  blks[0].gap = 255;
-  blks[0].blklen = 0xd1d2d3d4llu;
-  blks[1].gap = 1;
-  blks[1].blklen = 0xd1d2d3d4llu;
+  fr->ack.rangecnt = 2;
+  ranges = fr->ack.ranges;
+  ranges[0].gap = 255;
+  ranges[0].len = 0xd1d2d3d4llu;
+  ranges[1].gap = 1;
+  ranges[1].len = 0xd1d2d3d4llu;
   fr->ack.ecn.ect0 = 0;
   fr->ack.ecn.ect1 = 64;
   fr->ack.ecn.ce = 16384;
@@ -808,11 +822,11 @@ void test_ngtcp2_pkt_encode_ack_ecn_frame(void) {
   CU_ASSERT(fr->type == nfr->type);
   CU_ASSERT(fr->ack.largest_ack == nfr->ack.largest_ack);
   CU_ASSERT(fr->ack.ack_delay == nfr->ack.ack_delay);
-  CU_ASSERT(fr->ack.num_blks == nfr->ack.num_blks);
+  CU_ASSERT(fr->ack.rangecnt == nfr->ack.rangecnt);
 
-  for (i = 0; i < fr->ack.num_blks; ++i) {
-    CU_ASSERT(fr->ack.blks[i].gap == nfr->ack.blks[i].gap);
-    CU_ASSERT(fr->ack.blks[i].blklen == nfr->ack.blks[i].blklen);
+  for (i = 0; i < fr->ack.rangecnt; ++i) {
+    CU_ASSERT(fr->ack.ranges[i].gap == nfr->ack.ranges[i].gap);
+    CU_ASSERT(fr->ack.ranges[i].len == nfr->ack.ranges[i].len);
   }
 
   CU_ASSERT(fr->ack.ecn.ect0 == nfr->ack.ecn.ect0);
@@ -1244,7 +1258,7 @@ void test_ngtcp2_pkt_encode_retire_connection_id_frame(void) {
   fr.type = NGTCP2_FRAME_RETIRE_CONNECTION_ID;
   fr.retire_connection_id.seq = 1000000007;
 
-  framelen = 1 + ngtcp2_put_varint_len(fr.retire_connection_id.seq);
+  framelen = 1 + ngtcp2_put_uvarintlen(fr.retire_connection_id.seq);
 
   rv = ngtcp2_pkt_encode_retire_connection_id_frame(buf, sizeof(buf),
                                                     &fr.retire_connection_id);
@@ -1389,10 +1403,10 @@ void test_ngtcp2_pkt_validate_ack(void) {
   int rv;
   ngtcp2_ack fr;
 
-  /* too long first_ack_blklen */
+  /* too long first_ack_range */
   fr.largest_ack = 1;
-  fr.first_ack_blklen = 2;
-  fr.num_blks = 0;
+  fr.first_ack_range = 2;
+  fr.rangecnt = 0;
 
   rv = ngtcp2_pkt_validate_ack(&fr);
 
@@ -1400,21 +1414,21 @@ void test_ngtcp2_pkt_validate_ack(void) {
 
   /* gap is too large */
   fr.largest_ack = 250;
-  fr.first_ack_blklen = 1;
-  fr.num_blks = 1;
-  fr.blks[0].gap = 248;
-  fr.blks[0].blklen = 0;
+  fr.first_ack_range = 1;
+  fr.rangecnt = 1;
+  fr.ranges[0].gap = 248;
+  fr.ranges[0].len = 0;
 
   rv = ngtcp2_pkt_validate_ack(&fr);
 
   CU_ASSERT(NGTCP2_ERR_ACK_FRAME == rv);
 
-  /* too large blklen */
+  /* too large range len */
   fr.largest_ack = 250;
-  fr.first_ack_blklen = 0;
-  fr.num_blks = 1;
-  fr.blks[0].gap = 248;
-  fr.blks[0].blklen = 1;
+  fr.first_ack_range = 0;
+  fr.rangecnt = 1;
+  fr.ranges[0].gap = 248;
+  fr.ranges[0].len = 1;
 
   rv = ngtcp2_pkt_validate_ack(&fr);
 
@@ -1489,8 +1503,8 @@ void test_ngtcp2_pkt_write_retry(void) {
     token[i] = (uint8_t)i;
   }
 
-  spktlen = ngtcp2_pkt_write_retry(buf, sizeof(buf), NGTCP2_PROTO_VER_MAX,
-                                   &dcid, &scid, &odcid, token, sizeof(token),
+  spktlen = ngtcp2_pkt_write_retry(buf, sizeof(buf), NGTCP2_PROTO_VER_V1, &dcid,
+                                   &scid, &odcid, token, sizeof(token),
                                    null_retry_encrypt, &aead, &aead_ctx);
 
   CU_ASSERT(spktlen > 0);
@@ -1501,7 +1515,7 @@ void test_ngtcp2_pkt_write_retry(void) {
 
   CU_ASSERT(nread > 0);
   CU_ASSERT(NGTCP2_PKT_RETRY == nhd.type);
-  CU_ASSERT(NGTCP2_PROTO_VER_MAX == nhd.version);
+  CU_ASSERT(NGTCP2_PROTO_VER_V1 == nhd.version);
   CU_ASSERT(ngtcp2_cid_eq(&dcid, &nhd.dcid));
   CU_ASSERT(ngtcp2_cid_eq(&scid, &nhd.scid));
 
@@ -1517,19 +1531,20 @@ void test_ngtcp2_pkt_write_version_negotiation(void) {
   uint8_t buf[256];
   ngtcp2_ssize spktlen;
   const uint32_t sv[] = {0xf1f2f3f4, 0x1f2f3f4f};
-  uint8_t *p;
+  const uint8_t *p;
   size_t i;
   ngtcp2_cid dcid, scid;
+  uint32_t v;
 
   dcid_init(&dcid);
   scid_init(&scid);
 
   spktlen = ngtcp2_pkt_write_version_negotiation(
       buf, sizeof(buf), 133, dcid.data, dcid.datalen, scid.data, scid.datalen,
-      sv, arraylen(sv));
+      sv, ngtcp2_arraylen(sv));
 
   CU_ASSERT((ngtcp2_ssize)(1 + 4 + 1 + dcid.datalen + 1 + scid.datalen +
-                           arraylen(sv) * 4) == spktlen);
+                           ngtcp2_arraylen(sv) * 4) == spktlen);
 
   p = buf;
 
@@ -1537,9 +1552,9 @@ void test_ngtcp2_pkt_write_version_negotiation(void) {
 
   ++p;
 
-  CU_ASSERT(0 == ngtcp2_get_uint32(p));
+  p = ngtcp2_get_uint32(&v, p);
 
-  p += sizeof(uint32_t);
+  CU_ASSERT(0 == v);
 
   CU_ASSERT(dcid.datalen == *p);
 
@@ -1557,8 +1572,10 @@ void test_ngtcp2_pkt_write_version_negotiation(void) {
 
   p += scid.datalen;
 
-  for (i = 0; i < arraylen(sv); ++i, p += 4) {
-    CU_ASSERT(sv[i] == ngtcp2_get_uint32(p));
+  for (i = 0; i < ngtcp2_arraylen(sv); ++i) {
+    p = ngtcp2_get_uint32(&v, p);
+
+    CU_ASSERT(sv[i] == v);
   }
 }
 
